# -*- encoding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021/4/4 下午10:30
Written by Yuwei Jin(yuwei_jin@163.com)
"""
import torch
import torch.nn as nn

from sv.modules.packaged_conv import Conv1x1_BN_ReLU


class CPAMEnc(nn.Module):
    """
    CPAM encoding module
    """

    def __init__(self, in_channels, norm_layer=nn.BatchNorm2d):
        super(CPAMEnc, self).__init__()
        self.pool1 = nn.AdaptiveAvgPool2d(1)
        self.pool2 = nn.AdaptiveAvgPool2d(2)
        self.pool3 = nn.AdaptiveAvgPool2d(4)
        self.pool4 = nn.AdaptiveAvgPool2d(8)

        self.conv1 = Conv1x1_BN_ReLU(in_channels, in_channels)
        self.conv2 = Conv1x1_BN_ReLU(in_channels, in_channels)
        self.conv3 = Conv1x1_BN_ReLU(in_channels, in_channels)
        self.conv4 = Conv1x1_BN_ReLU(in_channels, in_channels)

    def forward(self, x):
        b, c, h, w = x.size()

        feat1 = self.conv1(self.pool1(x)).view(b, c, -1)
        feat2 = self.conv2(self.pool2(x)).view(b, c, -1)
        feat3 = self.conv3(self.pool3(x)).view(b, c, -1)
        feat4 = self.conv4(self.pool4(x)).view(b, c, -1)

        return torch.cat((feat1, feat2, feat3, feat4), 2)


class CPAMDec(nn.Module):
    """
    CPAM decoding module
    """

    def __init__(self, in_channels):
        super(CPAMDec, self).__init__()
        self.softmax = nn.Softmax(dim=-1)
        self.scale = nn.Parameter(torch.zeros(1))

        self.conv_query = nn.Conv2d(in_channels=in_channels, out_channels=in_channels // 4, kernel_size=1)  # query_conv2
        self.conv_key = nn.Linear(in_channels, in_channels // 4)  # key_conv2
        self.conv_value = nn.Linear(in_channels, in_channels)  # value2

    def forward(self, x, y):
        """
            inputs :
                x : input feature(N,C,H,W) y:gathering centers(N,K,M)
            returns :
                out : compact position attention feature
                attention map: (H*W)*M
        """
        m_batchsize, C, width, height = x.size()
        m_batchsize, K, M = y.size()

        proj_query = self.conv_query(x).view(m_batchsize, -1, width * height).permute(0, 2, 1)  # BxNxd
        proj_key = self.conv_key(y).view(m_batchsize, K, -1).permute(0, 2, 1)  # BxdxK
        energy = torch.bmm(proj_query, proj_key)  # BxNxK
        attention = self.softmax(energy)  # BxNxk

        proj_value = self.conv_value(y).permute(0, 2, 1)  # BxCxK
        out = torch.bmm(proj_value, attention.permute(0, 2, 1))  # BxCxN
        out = out.view(m_batchsize, C, width, height)
        out = self.scale * out + x
        return out


class DRAN(nn.Module):
    def __init__(self, in_chs):
        super(DRAN, self).__init__()
        self.pam_enc = CPAMEnc(in_chs)
        self.pam_dec = CPAMDec(in_chs)

    def forward(self, x):
        y = self.pam_enc(x).permute(0, 2, 1)
        y = self.pam_dec(x, y)
        return x


if __name__ == '__main__':
    x = torch.randn(1, 256, 32, 32)
    model = DRAN(256).eval()
    # y = model(x)

    from torchstat import stat
    stat(DRAN(256), input_size=(256, 32, 32))
    # input_sizes = [16, 32, 48, 64, 80, 96]
    # for s in input_sizes:
    #     pam = MSAM(in_chs=512).eval()
    #     print("size = ", s)
    #     stat(pam, input_size=(512, s, 2*s))




# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-10 20:10
Written by Yuwei Jin (642281525@qq.com)
"""

import torch
import torch.nn as nn

from sv.libs.tools.init_weight import init_weights


class SelfAttention(nn.Module):
    def __init__(self, in_chs, r=8):
        super(SelfAttention, self).__init__()

        query_chs = in_chs // r
        key_chs = in_chs // r

        self.query_conv = nn.Sequential(
            nn.Conv2d(in_chs, query_chs, kernel_size=1, bias=False),
            nn.BatchNorm2d(query_chs)
        )

        self.key_conv = nn.Sequential(
            nn.Conv2d(in_chs, key_chs, kernel_size=1, bias=False),
            nn.BatchNorm2d(key_chs)
        )

        self.value_conv = nn.Sequential(
            nn.Conv2d(in_chs, in_chs, kernel_size=1, bias=False),
            nn.BatchNorm2d(in_chs)
        )

        self.query_chs = query_chs
        self.key_chs = key_chs

        self.gamma = nn.Parameter(torch.zeros(1))

        init_weights(self.query_conv, self.key_conv, self.value_conv)

    def forward(self, x):
        B, C, H, W = x.size()

        q = self.query_conv(x).view(B, -1, H*W).permute(0, 2, 1)
        k = self.key_conv(x).view(B, -1, H*W)
        a = torch.bmm(q, k)
        a = torch.softmax(a, dim=-1)

        v = self.value_conv(x).view(B, -1, H*W)

        out = torch.bmm(v, a.permute(0, 2, 1))

        out = out.view(B, C, H, W)

        out = self.gamma * out + x

        return out


class PAM_Module(nn.Module):
    """ Position attention module"""

    # Ref from SAGAN
    def __init__(self, in_dim):
        super(PAM_Module, self).__init__()
        self.chanel_in = in_dim

        # 先经过3个卷积层生成3个新特征图B C D （尺寸不变）
        self.query_conv = nn.Conv2d(in_channels=in_dim, out_channels=in_dim // 8, kernel_size=1)
        self.key_conv = nn.Conv2d(in_channels=in_dim, out_channels=in_dim // 8, kernel_size=1)
        self.value_conv = nn.Conv2d(in_channels=in_dim, out_channels=in_dim, kernel_size=1)
        self.gamma = nn.Parameter(torch.zeros(1))  # α尺度系数初始化为0，并逐渐地学习分配到更大的权重

        self.softmax = nn.Softmax(dim=-1)  # 对每一行进行softmax

    def forward(self, x):
        """
            inputs :
                x : input feature maps( B × C × H × W)
            returns :
                out : attention value + input feature
                attention: B × (H×W) × (H×W)
        """
        m_batchsize, C, height, width = x.size()
        # B -> (N,C,HW) -> (N,HW,C)
        proj_query = self.query_conv(x).view(m_batchsize, -1, width * height).permute(0, 2, 1)
        # C -> (N,C,HW)
        proj_key = self.key_conv(x).view(m_batchsize, -1, width * height)
        # BC，空间注意图 -> (N,HW,HW)
        energy = torch.bmm(proj_query, proj_key)
        # S = softmax(BC) -> (N,HW,HW)
        attention = self.softmax(energy)
        # D -> (N,C,HW)
        proj_value = self.value_conv(x).view(m_batchsize, -1, width * height)
        # DS -> (N,C,HW)
        out = torch.bmm(proj_value, attention.permute(0, 2, 1))  # torch.bmm表示批次矩阵乘法
        # output -> (N,C,H,W)
        out = out.view(m_batchsize, C, height, width)

        out = self.gamma * out + x
        return out


if __name__ == '__main__':
    # SA = SelfAttention(512, r=8)
    SA = PAM_Module(512)
    from torchstat import stat

    input_sizes = [16, 32, 48, 64, 80, 96, 112]
    for s in input_sizes:
        print("size = ", s)
        stat(SA, input_size=(512, s, 2*s))



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-23 20:23 
Written by Yuwei Jin (642281525@qq.com)
"""
import math
import os
import random
import re
import shutil
import time

import cv2
import numpy as np
import pandas as pd
from PIL import Image
from torch import Tensor

from torch.nn import functional as F

# ---------------- file IO related utils ---------------------
import torch

from sv.libs.core.lr_scheduler import PolyLR, ExponentialLR, PolyLR, CosineLR, StepLR


def mk_dirs(path):
    if not os.path.exists(path):
        os.makedirs(path)


def mk_dirs_r(file_path):
    if os.path.exists(file_path):
        shutil.rmtree(file_path, ignore_errors=True)
    os.makedirs(file_path, exist_ok=True)


def rm_dir(path):
    if os.path.exists(path):
        shutil.rmtree(path)


def get_im_name_list(path, im_extensions='.tif'):
    """return example:
        ['1.tif', '2.tif']
    """
    out = []
    for f in os.listdir(path):
        if f.endswith(im_extensions):
            out.append(f)
    return out


def load_mean_std(cfg):
    """ loading mean and std config"""
    d = cfg.DATASET
    # load mean std
    if d.MEAN_STD == 'imagenet':
        mean = [0.485, 0.456, 0.406]  # imagenet mean and std value
        std = [0.229, 0.224, 0.225]
        return mean, std
    else:
        mean_std_file = os.path.join(d.TRAIN_SET, os.path.join(d.TRAIN_SET, d.NAME + '_mean_std.csv'))
        try:
            data = pd.read_csv(mean_std_file)
            mean = data['mean'].values.tolist()
            std = data['std'].values.tolist()
            return mean, std
        except FileNotFoundError:
            print('{} mean-std.csv file is not found.'.format(mean_std_file))


def save_train_cfg(cfg):
    import json
    save_dir = cfg.MODEL.OUTPUT_DIR
    mk_dirs_r(save_dir)
    file_name = "training_cfg.json"
    save_dir = os.path.join(save_dir, file_name)
    with open(save_dir, 'w+', encoding='utf-8') as f:
        json.dump(cfg, f, ensure_ascii=False)


def list_to_txt(str_list, output_name):
    with open(output_name, 'w') as f:
        for s in str_list:
            f.write(str(s) + '\n')


def txt_to_list(txt_file) -> list:
    out = []
    with open(txt_file, 'r') as f:
        lines = f.readlines()
        for s in lines:
            out.append(s.rstrip("\n"))
    return out


# ---------------- time consumption related utils ---------------------
class TimeCounter:
    def __init__(self, return_sec=False):
        self.start_time = 0
        self.end_time = 0

        self.return_sec = return_sec

    def tic(self):
        self.start_time = time.time()

    def toc(self):
        """default return seconds"""
        end_time = time.time()
        t = end_time - self.start_time
        t = round(t, 2)

        if self.return_sec:
            return t
        else:
            return t / 60


def time_counter(func):
    def inner(*args, **kwargs):
        start = time.time()
        func(*args, **kwargs)
        end = time.time()
        return (end - start) / 60.

    return inner


# ---------------- training, validate, and test related utils ---------------------
def fast_hist(input: np.ndarray, label: np.ndarray, num_class: int) -> np.ndarray:
    """calculation confusion matrix"""
    assert input.shape == label.shape

    input = input.flatten().astype('int')
    label = label.flatten().astype('int')

    mask = (label >= 0) & (label < num_class)  # 去除边界为0的情况
    label = num_class * label[mask] + input[mask]
    count = np.bincount(label, minlength=num_class ** 2)
    conf_mat = count.reshape(num_class, num_class)

    return conf_mat


# ---------------- tensor related utils ---------------------
def is_tensor(img):
    return isinstance(img, torch.Tensor)


def is_numpy_array(img):
    return isinstance(img, np.ndarray)


def tensor_to_numpy(data: torch.Tensor) -> np.ndarray:
    """Convert tensor data to numpy"""
    if is_tensor(data):
        data = data.squeeze().detach().cpu().numpy()
        return data
    else:
        raise TypeError('input data should be Tensor type. Got {}'.format(type(data)))


def tensor_binarize(prob: torch.Tensor, cfg) -> np.ndarray:
    if cfg.MODEL.OUTPUT_CHANNELS > 1:
        prob = torch.softmax(prob, dim=1)
        pred = torch.argmax(prob, dim=1)
    else:
        prob = torch.sigmoid(prob)
        pred = prob > 0.5

    pred = tensor_to_numpy(pred)

    return pred


# ---------------- training and validate utils ---------------------
def fixed_np_random_seed(seed=2048):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)


def fixed_torch_seed(cfg):
    seed = cfg.TRAIN.SEED
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    # torch.cuda.manual_seed_all(seed) # if you are using multi-GPU.
    if cfg.TRAIN.CUDNN_DETERMINISTIC:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def parse_lr_scheduler(cfg, optimizer, train_data):
    train_batch_steps = math.floor(len(train_data.dataset) / cfg.TRAIN.BATCH_SIZE)

    lr = cfg.TRAIN.LR
    warm_epoch = lr.WARMUP_EPOCH
    warm_start_lr = lr.WARMUP_START_LR

    if re.search(lr.SCHEDULER, 'ExponentialLR', re.IGNORECASE):
        scheduler = ExponentialLR(optimizer=optimizer, gamma=lr.POWER, eta_min_lr=lr.ETA_MIN_LR, warm_start_lr=warm_start_lr,
                                  warm_period=warm_epoch)
    elif re.search(lr.SCHEDULER, 'Poly', re.IGNORECASE):
        max_iters = cfg.TRAIN.EPOCHS * train_batch_steps
        warmup_period = warm_epoch * train_batch_steps
        scheduler = PolyLR(optimizer=optimizer, power=lr.POWER, max_iterations=max_iters, warm_start_lr=warm_start_lr,
                           warm_period=warmup_period)

    elif re.search(lr.SCHEDULER, 'Cosine', re.IGNORECASE):
        max_iters = cfg.TRAIN.EPOCHS * train_batch_steps
        warmup_period = lr.WARMUP_STEPS * train_batch_steps
        scheduler = CosineLR(optimizer=optimizer, eta_min_lr=lr.ETA_MIN_LR, T_max=max_iters, warm_start_lr=warm_start_lr,
                             warm_period=warmup_period)
    elif re.search(lr.SCHEDULER, 'Step', re.IGNORECASE):
        scheduler = StepLR(optimizer=optimizer, warm_start_lr=warm_start_lr, step_size=lr.STEP_SIZE, warm_period=warm_epoch,
                           gamma=lr.POWER)
    else:
        raise NotImplementedError("{} is not implemented.".format(lr.SCHEDULER))

    return scheduler


# ---------------- saving utils ---------------------
def save_binary_map(prob: torch.Tensor, im_name: str, save_dir: str, cfg):
    res = tensor_binarize(prob) * 255  # res
    filename = os.path.join(save_dir, cfg.MODEL.NAME + '_' + im_name + '_binary.png')
    cv2.imwrite(filename, res)


def save_prob_map(im_name, prob, cfg, save_dir=None):
    if is_tensor(prob):
        if cfg.MODEL.OUTPUT_CHANNELS > 1:
            prob = torch.softmax(prob, dim=1)
            prob = tensor_to_numpy(prob)[1, :, :]
        else:
            prob = torch.sigmoid(prob)
            # prob = torch.relu(prob)
            prob = tensor_to_numpy(data=prob)
            # eps = 1e-10
            # prob = prob / (prob.max() + eps)
            # prob = np.where(prob<0, 0, prob)

    prob = np.round(prob * 255)

    save_dir = cfg.TEST.PROB_SAVE_DIR if save_dir is None else save_dir
    filename = os.path.join(save_dir, cfg.MODEL.NAME + '_' + im_name + '_prob.png')
    cv2.imwrite(filename, prob.astype(np.uint8))


def save_palette_map(im_name, prob, gt, cfg):
    # obtain binary prediction
    if isinstance(prob, np.ndarray):
        p = prob > 0.5
    else:
        p = tensor_binarize(prob)

    def inner_fun(p_flg, gt_flag, color):
        r = np.where((p == p_flg) & (gt == gt_flag), color[0], 0)
        g = np.where((p == p_flg) & (gt == gt_flag), color[1], 0)
        b = np.where((p == p_flg) & (gt == gt_flag), color[2], 0)
        return cv2.merge([b, g, r]).astype(np.uint8)

    bg = inner_fun(0, 0, cfg.TEST.BACKGROUND_COLOR)
    tp = inner_fun(1, 1, cfg.TEST.TP_COLOR)
    fp = inner_fun(1, 0, cfg.TEST.FP_COLOR)
    fn = inner_fun(0, 1, cfg.TEST.FN_COLOR)

    out = (tp + bg + fp + fn)

    save_dir = cfg.TEST.PROB_SAVE_DIR
    output_path = os.path.join(save_dir, cfg.MODEL.NAME + '_' + im_name + '_palette.png')
    cv2.imwrite(output_path, out)


def save_alpha_map(im_name, prob, gt, cfg, save_dir=None):
    data = cfg.DATASET
    save_dir = cfg.TEST.PROB_SAVE_DIR
    if cfg.TEST.STITCH_RES:
        ori_image = Image.open(os.path.join(data.TEST_WHOLE_IMAGE, im_name + data.IMG_EXTENSION)).convert('RGBA')
    else:
        ori_image = Image.open(os.path.join(data.TEST_SET, 'image', im_name + data.IMG_EXTENSION)).convert('RGBA')

    if not is_tensor(prob):
        p = prob > 0.5
    else:
        p = tensor_binarize(prob)

    def inner_fun(p_flg, gt_flag, color):
        r = np.where((p == p_flg) & (gt == gt_flag), color[0], 0)
        g = np.where((p == p_flg) & (gt == gt_flag), color[1], 0)
        b = np.where((p == p_flg) & (gt == gt_flag), color[2], 0)
        return cv2.merge([r, g, b]).astype(np.uint8)

    tp = inner_fun(1, 1, cfg.TEST.TP_COLOR)
    fp = inner_fun(1, 0, cfg.TEST.FP_COLOR)
    fn = inner_fun(0, 1, cfg.TEST.FN_COLOR)

    mask = (tp + fp + fn)
    mask = Image.fromarray(mask).convert('RGBA')
    image = Image.blend(ori_image, mask, cfg.TEST.ALPHA)

    # image_rgba = np.array(image)
    # ori_image = np.array(ori_image)
    # #
    # ori_image[:, :, 0] = np.where(p[:, :] == 1, image_rgba[:, :, 0], ori_image[:, :, 0])
    # ori_image[:, :, 1] = np.where(p[:, :] == 1, image_rgba[:, :, 1], ori_image[:, :, 1])
    # ori_image[:, :, 2] = np.where(p[:, :] == 1, image_rgba[:, :, 2], ori_image[:, :, 2])
    # #
    # image = Image.fromarray(ori_image.astype(np.uint8))

    save_dir = cfg.TEST.PROB_SAVE_DIR if save_dir is None else save_dir
    image.save(os.path.join(save_dir, cfg.MODEL.NAME + '_' + im_name + '_alpha.png'))


def edge_extraction(x: torch.Tensor, operator='laplacian') -> torch.Tensor:
    """extract the edge of feature map by utilizing specific kernel operator
    Args:
        x: input feature map (type: torch.Tensor)
        operator: 'laplacian', 'sobel', 'prewitt', default is laplacian
    Return:
        soft edge
    """

    def _conv2d(input, kernel):
        kernel.requires_grad = False
        return F.conv2d(input, weight=kernel, padding=1, stride=1, bias=None)

    if operator == 'laplacian':
        kernel = -1 * torch.ones(1, 1, 3, 3)
        kernel[:, :, 1, 1] = 8.
        return _conv2d(x, kernel)
    elif operator == 'sobel':
        kernel_x = torch.Tensor([[[[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]]]])
        gx = _conv2d(input, kernel_x)
        kernel_y = torch.Tensor([[[[1, 2, 1], [0, 0, 0], [-1, -2, -1]]]])
        gy = _conv2d(input, kernel_y)
        return torch.sqrt(torch.pow(gx, 2) + torch.pow(gy, 2))
    elif operator == 'prewitt':
        kernel_x = torch.Tensor([[[[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]]]])
        gx = _conv2d(input, kernel_x)
        kernel_y = torch.Tensor([[[[1, 1, 1], [0, 0, 0], [-1, -1, -1]]]])
        gy = _conv2d(input, kernel_y)
        return torch.sqrt(torch.pow(gx, 2) + torch.pow(gy, 2))
    else:
        raise NotImplementedError


def channel_shuffle(x, groups):
    batch_size, num_channels, height, width = x.data.size()

    channels_per_group = num_channels // groups

    # reshape
    x = x.view(batch_size, groups, channels_per_group, height, width)

    # transpose
    # - contiguous() required if transpose() is used before view().
    #   See https://github.com/pytorch/pytorch/issues/764
    x = torch.transpose(x, 1, 2).contiguous()

    # flatten
    x = x.view(batch_size, -1, height, width)

    return x


def global_weighted_average_pool(x):
    eps = 1e-10
    b, c, _, _ = x.size()
    x1 = x.view(b, c, -1)
    w = x / (torch.sum(x1, dim=2).view(b, c, 1, 1) + eps)
    y = x * w
    y = y.view(b, c, -1)
    y = torch.sum(y, dim=2).view(b, c, 1, 1)


def cal_fps(model, input_size=(3, 512, 512), epochs=200):
    tic = torch.cuda.Event(enable_timing=True)
    toc = torch.cuda.Event(enable_timing=True)

    times = np.zeros((epochs, 1))

    model = model.cuda().eval()
    x = torch.randn(input_size).unsqueeze(dim=0).cuda()

    # GPU-WARM-UP
    with torch.no_grad():
        for _ in range(20):
            _ = model(x)

    # MEASURE PERFORMANCE
    mean_time, fps = [], []
    for _ in range(3):
        with torch.no_grad():
            for epoch in range(epochs):
                tic.record()
                _ = model(x)
                toc.record()
                # WAIT FOR GPU SYNC
                torch.cuda.synchronize()
                times[epoch] = tic.elapsed_time(toc)

        temp = np.sum(times) / epochs
        mean_time.append(temp)
        fps.append(1000.0 / temp)
    print('mean: {:.3f}ms, fps: {:.2f}'.format(np.mean(mean_time), np.mean(fps)))


def get_distance_transform(label: np.ndarray) -> Tensor:
    """binary label distance transform"""
    label = np.uint8(label)
    dist = cv2.distanceTransform(label, cv2.DIST_L2, 0)
    eps = 1e-10
    dist = dist / (dist.max() + eps)

    dist = np.where((0.3 > dist) & (dist > 0), 1, 0)

    # plt.imshow(dist)
    # plt.show()

    # convert to tensor
    dist = torch.from_numpy(dist).unsqueeze(dim=0).cuda()

    return dist


def add_list_to_dataframe(data, index, df):
    """Convert list to DataFrame"""
    columns = df.columns
    dic = dict(map(lambda x, y: [x, y], columns, data))
    df.loc[index] = dic

    return df



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-24 17:03 
Written by Yuwei Jin (642281525@qq.com)
"""

import torch
import torch.nn as nn

from sv.modules.seg_modules import UNetEncBlock, Bottleneck, DoubleConv, UNetDecBlock, Classifier, upsample
from sv.modules.backbones.resnet.ori_resnet import resnet18

supported_encoder = ['naive', 'resnet-18d', 'resnet-18', 'vgg-16', 'resnet-34', 'resnet-50', 'resnet-101']


class _NaiveUNetEnc(nn.Module):
    def __init__(self):
        super(_NaiveUNetEnc, self).__init__()
        self.enc1 = UNetEncBlock(3, 64)
        self.enc2 = UNetEncBlock(64, 128)
        self.enc3 = UNetEncBlock(128, 256)
        self.enc4 = UNetEncBlock(256, 512)
        self.bridge_conv = DoubleConv(512, 1024)

    def forward(self, x):
        x1 = self.enc1(x)
        x2 = self.enc2(x1[1])
        x3 = self.enc3(x2[1])
        x4 = self.enc4(x3[1])
        x5 = self.bridge_conv(x4[1])
        return x1[0], x2[0], x3[0], x4[0], x5


class _UNetDec(nn.Module):
    def __init__(self, num_class, in_chs_list=[64, 128, 256, 512, 1024], use_ds=False):
        super(_UNetDec, self).__init__()

        self.dec = nn.ModuleList()

        for i in range(len(in_chs_list) -1, 0, -1):
            m = UNetDecBlock(in_chs_list[i-1], in_chs_list[i])
            self.dec.append(m)

        self.cls = Classifier(in_chs_list[0], num_class)

    def forward(self, feats):
        feats = list(feats)
        feats.reverse()
        for i, m in enumerate(self.dec):
            low_f = feats[i+1]
            if i == 0:
                x = m(low_f, feats[i])
            else:
                x = m(low_f, x)

        p = self.cls(x)

        return p


class UNet(nn.Module):
    def __init__(self, num_class=1, encoder_type='naive'):
        super().__init__()

        if encoder_type not in supported_encoder:
            raise ValueError('encoder type is not supported. Got {}'.format(encoder_type))

        if encoder_type == 'naive':
            self.encoder = _NaiveUNetEnc()
            self.decoder = _UNetDec(num_class, [64, 128, 256, 512, 1024])

        if encoder_type == 'resnet-18':
            self.encoder = resnet18(pretrained=True)
            self.decoder = _UNetDec(num_class, [64, 128, 256, 512])

    def forward(self, x):
        feats = self.encoder(x)
        p = self.decoder(feats)
        p = upsample(p, x.size()[2:])
        return p


if __name__ == '__main__':
    model = UNet(encoder_type='resnet-18').eval()
    x = torch.randn(1, 3, 512, 512)
    print(model(x).size())



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-23 21:43 
Written by Yuwei Jin (642281525@qq.com)
"""

import os
import cv2
import math
import json

import numpy as np
from PIL import Image
from multiprocessing import Pool

from sv.libs.tools import utils


class SlideWinCrop:
    """Crop a big image and corresponding label into small patches
        mainly used for inferring.
    """

    def __init__(self, crop_cfg, use_pil_read=False):
        self.patch_size = crop_cfg['patch_size']
        self.win = crop_cfg['win_size']
        self.im_suffix = crop_cfg['im_suffix']
        self.gt_suffix = crop_cfg['gt_suffix']
        self.dataset_cfg = crop_cfg['dataset_cfg']
        self.dataset_name = crop_cfg['dataset_name']
        self.phase = crop_cfg['phase']

        root_dir = crop_cfg['root_dir']

        output_root_dir = crop_cfg['output_root_dir']
        output_root_dir = os.path.join(output_root_dir, 'test')

        self.im_src = os.path.join(root_dir, 'image')
        self.gt_src = os.path.join(root_dir, 'label')

        self.im_output_src = os.path.join(output_root_dir, 'image')
        self.gt_output_src = os.path.join(output_root_dir, 'label')
        self.crop_info_dir = os.path.join(output_root_dir, 'crop_info')

        self.use_pil_read = use_pil_read



# -*- coding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021-01-01 10:56
Written by Yuwei Jin (642281525@qq.com)
"""

"""
# ResNet-D backbone with deep-stem
# Code Adapted from:
# https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py
"""
import torch
import torch.nn as nn
import torch.utils.model_zoo as model_zoo

__all__ = ['ResNet', 'resnet18']

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class BasicBlock(nn.Module):
    """
    Basic Block for Resnet
    """
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    """
    Bottleneck Layer for Resnet
    """
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * self.expansion, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    """
    Resnet
    """

    def __init__(self, block, layers, num_classes=1000, os=16):
        self.inplanes = 128
        super(ResNet, self).__init__()
        self.conv1 = nn.Sequential(
            conv3x3(3, 64, stride=2),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            conv3x3(64, 64),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            conv3x3(64, 128))
        self.bn1 = nn.BatchNorm2d(128)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        if os == 16:
            self.layer4 = self._make_layer(block, 512, layers[3], stride=1)
        else:
            self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = nn.AvgPool2d(7, stride=1)
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for index in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x1 = self.layer1(x)
        x2 = self.layer2(x1)
        x3 = self.layer3(x2)
        x4 = self.layer4(x3)

        return x1, x2, x3, x4


def resnet18(pretrained=True, src=None, **kwargs):
    """Constructs a ResNet-18 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        :param pretrained:
        :param src:
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        # import os
        # src = os.getcwd()
        # src = os.path.join(src, 'resnet18-deep-inplane128.pth')
        model.load_state_dict(torch.load(src, map_location="cpu"))
    return model


if __name__ == '__main__':
    model =  resnet18(pretrained=True)
    model(torch.randn(1, 3, 512, 512))


# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-06-22 09:17
Written by Yuwei Jin (642281525@qq.com)
"""

import random

import cv2
import numpy as np
import torch
from PIL import Image, ImageFilter
from torchvision.transforms import functional as F


def _func_wrap(images, func, **kwargs):
    out = [func(im, **kwargs) for im in images]
    return out


def _to_tensor(im, gt):
    if not isinstance(im, torch.Tensor) or not isinstance(gt, torch.Tensor):
        if isinstance(im, np.ndarray):
            im = torch.from_numpy(im).permute(2, 0, 1).float()
            gt = torch.from_numpy(gt).unsqueeze(dim=0)

        if isinstance(im, Image.Image):
            im, gt = _func_wrap([im, gt], F.to_tensor)

    return im, gt


class Compose(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, im, gt=None):
        if gt is not None:
            for t in self.transforms:
                im, gt = t(im, gt)
            return im, gt
        else:
            for t in self.transforms:
                im = t(im)
            return im


class RandomScale(object):
    def __init__(self, cfg):
        self.p = cfg.TRAIN.AUG.MS_P
        self.scale_list = cfg.TRAIN.AUG.SCALE_LIST

    def __call__(self, im: np.ndarray, gt: np.ndarray):
        if random.random() < self.p:
            scale = float(random.choice(self.scale_list))
            if scale != 1.0:
                h, w = im.shape[0], im.shape[1]
                new_h = int(scale * h + 0.5)
                new_w = int(scale * w + 0.5)

                im = cv2.resize(im, dsize=(new_h, new_w), interpolation=cv2.INTER_LINEAR)
                gt = cv2.resize(gt, dsize=(new_h, new_w), interpolation=cv2.INTER_NEAREST)

        return im, gt


class RandomCrop(object):
    def __init__(self, cfg, mean):
        self.crop_size = cfg.TRAIN.AUG.CROP_SIZE,
        self.mean = None,
        self.center_crop = cfg.TRAIN.AUG.CENTER_CROP,
        self.ignore_label = cfg.DATASET.IGNORE_LABEL

        self.mean = mean

    def __call__(self, im: np.ndarray, gt: np.ndarray):
        h, w = im.shape[0], im.shape[1]

        pad_h = max(self.crop_size[0] - h, 0)
        pad_w = max(self.crop_size[1] - w, 0)

        if pad_h > 0 or pad_w > 0:
            pad_val = np.asarray(self.mean)
            im = cv2.copyMakeBorder(im, 0, pad_h, 0, pad_w, cv2.BORDER_CONSTANT, value=pad_val)
            gt = cv2.copyMakeBorder(gt, 0, pad_h, 0, pad_w, cv2.BORDER_CONSTANT, value=self.ignore_label)
            x1 = 0
            y1 = 0
        elif pad_w == 0 or pad_w == 0:
            x1, y1 = 0, 0
        else:
            x1 = np.random.randint(0, abs(h - self.crop_size[0]))
            y1 = np.random.randint(0, abs(w - self.crop_size[1]))

        x2 = x1 + self.crop_size[0]
        y2 = y1 + self.crop_size[1]

        im = im[x1:x2, y1:y2, :]
        gt = gt[x1:x2, y1:y2]

        return im, gt


class RandomFlip(object):
    def __init__(self, cfg):
        self.p = cfg.TRAIN.AUG.FLIP_P

    def __call__(self, im, gt):
        if random.random() < self.p:
            if np.random.randint(2) == 1:  # horizontally flip
                flipcode = 1
            else:
                flipcode = 0
            im, gt = cv2.flip(im, flipCode=flipcode), cv2.flip(gt, flipCode=flipcode)
        return im, gt


class RandomRotation(object):
    def __init__(self, cfg):
        self.angle = cfg.TRAIN.AUG.ROTATION_ANGLE
        self.p = cfg.TRAIN.AUG.ROTATION_P

    def __call__(self, im, gt):
        im, gt = _to_tensor(im, gt)

        if random.random() < self.p:
            if isinstance(self.angle, int):
                angle = random.uniform(-1 * self.angle, self.angle)
            else:
                angle = np.random.choice(self.angle)

            angle = float(angle)

            im = F.rotate(im, angle=angle, interpolation=F.InterpolationMode.BILINEAR)
            gt = F.rotate(gt, angle=angle, interpolation=F.InterpolationMode.NEAREST)

        return im, gt


class RandomColorJitter(object):
    def __init__(self, cfg):
        brightness_factor = cfg.TRAIN.AUG.BRIGHTNESS_FACTOR,
        contrast_factor = cfg.TRAIN.AUG.CONTRAST_FACTOR,
        hue_factor = cfg.TRAIN.AUG.HUE_FACTOR,
        saturation_factor = cfg.TRAIN.AUG.SATURATION_FACTOR,
        p = cfg.TRAIN.AUG.C0LOR_JITTER_P

        if not (0 <= hue_factor <= 0.5):
            raise ValueError('hue_factor is not in [-0.5, 0.5].'.format(hue_factor))
        self.p = p
        self.brightness_factor = random.uniform(max(0, 1 - brightness_factor), 1 + brightness_factor)
        self.contrast_factor = random.uniform(max(0, 1 - contrast_factor), 1 + contrast_factor)
        self.hue_factor = random.uniform(-1 * hue_factor, hue_factor)
        self.saturation_factor = random.uniform(max(0, 1 - saturation_factor), 1 + saturation_factor)

    def __call__(self, im, gt):
        im, gt = _to_tensor(im, gt)

        if random.random() < self.p:
            if np.random.randint(4) == 0:
                im = F.adjust_hue(im, hue_factor=self.hue_factor)
            elif np.random.randint(4) == 1:
                im = F.adjust_saturation(im, saturation_factor=self.saturation_factor)
            elif np.random.randint(4) == 2:
                im = F.adjust_brightness(im, brightness_factor=self.brightness_factor)
            else:
                im = F.adjust_contrast(im, contrast_factor=self.contrast_factor)

        return im, gt


class RandomGaussianBlur(object):
    def __init__(self, cfg):
        self.radius = cfg.TRAIN.AUG.GAUSSIAN_BLUR_RADIUS
        self.p = cfg.TRAIN.AUG.GAUSSIAN_BLUR_P

    def _blur(self, img):
        if not isinstance(img, Image.Image):
            if isinstance(img, torch.Tensor):
                img = F.to_pil_image(img)
            else:
                img = Image.fromarray(img)

        x = img.filter(ImageFilter.GaussianBlur(radius=self.radius))
        x = F.to_tensor(x)
        return x

    def __call__(self, im, gt):
        if random.random() < self.p:
            im = self._blur(im)
        return im, gt


class RandomGaussianNoise(object):
    def __init__(self, cfg):
        self.mean = cfg.TRAIN.AUG.GAUSSIAN_NOISE_MEAN
        self.var = cfg.TRAIN.AUG.GAUSSIAN_NOISE_VAR
        self.p = cfg.TRAIN.AUG.GAUSSIAN_NOISE_P

    def _add_noise(self, img):
        img = np.asarray(img)
        img = np.array(img / 255, dtype=float)
        noise = np.random.normal(self.mean, self.var ** 0.5, img.shape)
        out = img + noise

        if out.min() < 0:
            low_clip = -1
        else:
            low_clip = 0

        out = np.clip(out, low_clip, 1.0)
        out = np.uint8(out * 255)
        out = Image.fromarray(out).convert('RGB')
        return out

    def __call__(self, im, gt):
        im, gt = _to_tensor(im, gt)
        if random.random() < self.p:
            im = self._add_noise(im)
        return im, gt


class Normalize(object):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, im, gt=None):
        if not isinstance(im, torch.Tensor):
            im = torch.from_numpy(im).permute(2, 0, 1).float()
            gt = torch.from_numpy(gt).unsqueeze(dim=0)

        im = F.normalize(im, mean=self.mean, std=self.std)

        if gt is not None:
            return im, gt
        else:
            return im


if __name__ == '__main__':

    def imshow(t0_im, t1_im, label):
        # t0_im = np.uint8(t0_im)
        # t1_im = np.uint8(t1_im)

        from matplotlib import pyplot as plt
        plt.subplot(131)
        plt.imshow(t0_im)
        plt.subplot(132)
        plt.imshow(t1_im)
        plt.subplot(133)
        plt.imshow(label)
        plt.show()


    def tensor2np(image):
        if len(image.shape) > 2:
            return image.permute(1, 2, 0).cpu().detach().numpy()
        else:
            return image.cpu().detach().numpy()


    im = cv2.imread('69_im.tif')
    im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB) / 255.0
    gt = cv2.imread('69.tif', cv2.IMREAD_GRAYSCALE)

    tf = Compose([
        RandomScale(scale_list=[0.5], p=1),
        RandomCrop(crop_size=[352, 352], mean=[0.485, 0.456, 0.406], center_crop=False, ignore_label=0),
        RandomFlip(p=1),
        RandomRotation(angle=[90, 180, 270], p=1),
        RandomColorJitter(p=1),
        Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        # RandomGaussianBlur(p=1),
        # RandomGaussianNoise(p=1)
    ])

    im, gt = tf(im, gt)

    im = im.permute(1, 2, 0)
    im = utils.tensor_to_numpy(im)
    gt = utils.tensor_to_numpy(gt)

    from matplotlib import pyplot as plt

    plt.subplot(1, 2, 1)
    plt.imshow(im)
    plt.subplot(1, 2, 2)
    plt.imshow(gt)
    plt.show()

    # tf = Compose([
    #     RandomFlip(p=1),
    #     RandomRotation(p=1),
    #     # RandomGaussianNoise(p=1),
    #     # RandomGaussianBlur(p=1),
    #     RandomColorJitter(p=1),
    #     ToTensor(),
    #     Normalize(means={
    #         't0': [0.485, 0.456, 0.406],
    #         't1': [0.485, 0.456, 0.406]
    #     },
    #         stds={
    #             't0': [0.229, 0.224, 0.225],
    #             't1': [0.229, 0.224, 0.225]
    #         }
    #     ),
    #     RandomScale(p=1, scale_list=[0.5])
    # ])
    #
    # t0_im, t1_im, label = tf(ori_t0_im, ori_t1_im, ori_label)
    # t0_im, t1_im, label = _func_wrap([t0_im, t1_im, label], tensor2np)
    # imshow(t0_im, t1_im, label)



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-06-11 15:36
Written by Yuwei Jin (642281525@qq.com)
"""
import os
import warnings
import cv2
import torch
import random
import numpy as np

from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset

from sv.libs.data.dataset import joint_transforms
from sv.libs.tools import utils


def _parse_train_aug_method(cfg, mean):
    aug = cfg.TRAIN.AUG
    out = []
    if aug.MULTI_SCALE:
        out.append(joint_transforms.RandomScale(cfg))

    out.append(joint_transforms.RandomCrop(cfg, mean))

    if aug.FLIP:
        out.append(joint_transforms.RandomFlip(cfg))
    if aug.ROTATION:
        out.append(joint_transforms.RandomRotation(cfg))
    if aug.COLOR_JITTER:
        out.append(joint_transforms.RandomColorJitter(cfg))
    if aug.GAUSSIAN_BLUR:
        out.append(joint_transforms.RandomGaussianBlur(cfg))
    if aug.GAUSSIAN_NOISE:
        out.append(joint_transforms.RandomGaussianNoise(cfg))
    return out


def _read_image(im_src, gt_src, gt_normalize=True):
    """ for two-class semantic segmentation task, set gt_normalize=True"""
    im = cv2.imread(im_src)
    im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB) / 255.0
    gt = cv2.imread(gt_src, cv2.IMREAD_GRAYSCALE)

    if gt_normalize:
        gt /= 255

    return im, gt


class _TrainValTestDataset(Dataset):
    def __init__(self, cfg, training=True, test=False):

        if cfg.DATASET.NUM_CLASS > 2:
            warnings.warn('the label will be normalized to {0, 1}')

        # load mean_std
        mean, std = utils.load_mean_std(cfg)

        # define data augmentation transformation
        temp = joint_transforms.Normalize(mean, std)
        if training:
            transforms = _parse_train_aug_method(cfg, mean)
            transforms.append(temp)
        else:
            transforms = [temp]
        self.transforms = joint_transforms.Compose(transforms)

        # define data set dir
        if training and test:  # validation phase
            dataset_dir = cfg.DATASET.VALID_SET
        elif training and not test:  # training phase
            dataset_dir = cfg.DATASET.TRAIN_SET
        else:
            dataset_dir = cfg.DATASET.TEST_SET

        # generate samples
        self.samples = self._make_samples(dataset_dir=dataset_dir)

        self.test = test
        self.cfg = cfg

    def _make_samples(self, dataset_dir):
        im_dir = os.path.join(dataset_dir, 'image')
        gt_dir = os.path.join(dataset_dir, 'label')

        samples = []
        for f in os.listdir(im_dir):
            if f.endswith(self.cfg.DATASET.IMG_EXTENSION):
                im = os.path.join(im_dir, f)
                gt = os.path.join(gt_dir, f)
                samples.append((im, gt))

        if len(samples) == 0:
            raise FileNotFoundError('None of image sample is found.')

        return samples

    def __getitem__(self, index):
        im_src, gt_src = self.samples[index]

        im, gt = _read_image(im_src, gt_src)

        if self.test:
            f = im_src.split('/')
            im_name = f[-1]
            im, gt = self.transforms(im, gt)
            return im_name, im, gt
        else:
            im, gt = self.transforms(im, gt)
            return im, gt

    def __len__(self):
        return len(self.samples)


def _seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2 ** 32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


def load_train_data(cfg):
    dataset = _TrainValTestDataset(cfg, training=True, test=False)

    time_counter = utils.TimeCounter(return_sec=True)
    time_counter.tic()
    data = DataLoader(
        dataset=dataset,
        batch_size=cfg.TRAIN.BATCH_SIZE,
        shuffle=True,
        drop_last=cfg.DATALOADER.DROP_LAST,
        num_workers=cfg.DATALOADER.WORKERS,
        pin_memory=cfg.DATALOADER.PIP_MEMORY,
        worker_init_fn=_seed_worker
    )
    t = time_counter.toc()

    print('{} training images were loaded from dir:{}'.format(dataset.__len__(), cfg.DATASET.TRAIN_SET))
    print('elapsed time for loading training set is {} seconds'.format(t))

    return data


def load_val_data(cfg):
    dataset = _TrainValTestDataset(cfg, training=False, test=False)

    time_counter = utils.TimeCounter(return_sec=True)
    time_counter.tic()
    data = DataLoader(
        dataset=dataset,
        batch_size=cfg.TRAIN.BATCH_SIZE,
        shuffle=True,
        num_workers=cfg.DATALOADER.WORKERS,
        drop_last=True,
        pin_memory=True
    )
    t = time_counter.toc()

    print('{} validate images were loaded from dir:{}'.format(dataset.__len__(), cfg.DATASET.VALID_SET))
    print('elapsed time for loading validate set is {} seconds'.format(t))

    return data


def load_test_data(cfg):
    dataset = _TrainValTestDataset(cfg, training=False, test=True)

    time_counter = utils.TimeCounter(return_sec=True)
    time_counter.tic()
    data = DataLoader(
        dataset=dataset,
        batch_size=1,
        shuffle=False,
        num_workers=1,
        drop_last=False,
        pin_memory=True
    )

    t = time_counter.toc()

    print('{} test images were loaded from dir:{}'.format(dataset.__len__(), cfg.DATASET.VALID_SET))
    print('elapsed time for loading test set is {} seconds'.format(t))

    return data



# -*- encoding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021/4/27 下午7:14
Written by Yuwei Jin (yuwei_jin@163.com)
"""
from functools import reduce

import torch
import torch.nn as nn

from sv.modules.packaged_conv import Conv3x3_BN_ReLU, Conv1x1_BN_ReLU, Conv1x1


class SKModule(nn.Module):
    def __init__(self, in_chs, out_chs, reduction_dim=16, num_branch=2, max_l=32):
        super(SKModule, self).__init__()

        d = max(in_chs // reduction_dim, max_l)  # 计算向量Z 的长度d
        self.num_branch = num_branch
        self.out_chs = out_chs

        self.conv = nn.ModuleList()
        for i in range(num_branch):
            self.conv.append(Conv3x3_BN_ReLU(2*in_chs, out_chs, dilation=i + 1, padding=i + 1))

        self.fc = nn.Sequential(
            Conv1x1_BN_ReLU(out_chs, d),
            Conv1x1(d, out_chs * num_branch, bias=False)
        )

        self.global_avg_pool = nn.AdaptiveAvgPool2d(1)

    def forward(self, x):
        b, c, h, w = x.size()

        # different scale conv
        conv_out = []
        for i, conv in enumerate(self.conv):
            conv_out.append(conv(x))

        U = reduce(lambda v0,v1:v0+v1,conv_out)
        p = self.global_avg_pool(U)
        p = self.fc(p)
        p = p.view(b, self.num_branch, self.out_chs, -1)
        a = torch.softmax(p, dim=1)

        a = list(a.chunk(self.num_branch, dim=1))  # split to a and b   chunk为pytorch方法，将tensor按照指定维度切分成 几个tensor块
        a = list(map(lambda x: x.reshape(b, self.out_chs, 1, 1), a))  # 将所有分块  调整形状，即扩展两维
        V = list(map(lambda x, y: x * y, conv_out, a))  # 权重与对应  不同卷积核输出的U 逐元素相乘
        # V = reduce(lambda x, y: x + y, V)  # 两个加权后的特征 逐元素相加
        out = torch.cat(V, dim=1)

        return out


if __name__ == '__main__':
    x = torch.rand(1, 256, 256, 256)
    y = torch.rand(1, 256, 256, 256)

    model = SKModule(256, 256).eval()

    out = model(x, y)

    print(out.shape)



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-18 20:53 
Written by Yuwei Jin (642281525@qq.com)
"""
import os
import re

import torch
import numpy as np
from tensorboardX import SummaryWriter
from abc import ABC, abstractmethod
from nb_log import get_logger
from torch.cuda.amp import autocast, GradScaler

from sv.libs.tools import utils

scaler = GradScaler()  # fp-16 training

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


def _find_best_value_index(value_list: list, higher_better=True):
    def _sub_func(value: list):
        temp_list = []
        for i, v in enumerate(value_list):
            if v == value:
                temp_list.append(i)
        return temp_list[-1]

    if higher_better:
        max_val = max(value_list)
        return _sub_func(value=max_val)
    else:
        min_val = min(value_list)
        return _sub_func(value=min_val)


class _BaseTrainer(ABC):
    """Base class for trainer"""

    def __init__(self, cfg, model, optimizer, scheduler, criterion, evaluator):
        self.curr_epoch = 1
        self.curr_step = 1
        self.curr_saved_epoch = None

        self.train_batch_loss = 0
        self.valid_batch_loss = 0

        self.key_metric_record = []
        self.last_best_epoch = 0

        self.num_class = cfg.DATASET.NUM_CLASS

        self.train_batch_hist = np.zeros((self.num_class,) * 2)
        self.valid_batch_hist = np.zeros((self.num_class,) * 2)

        self.employed_metrics = ['loss']
        self.employed_metrics.extend(cfg.TRAIN.EMPLOYED_METRICS)

        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.criterion = criterion
        self.evaluator = evaluator

        self.writer = SummaryWriter(cfg.TRAIN.LOGS_SAVE_DIR)

        self.total_time = 0

        self.cfg = cfg

    def epoch_reset(self):
        self.train_batch_hist = np.zeros((self.num_class,) * 2)
        self.valid_batch_hist = np.zeros((self.num_class,) * 2)
        self.train_batch_loss = 0
        self.valid_batch_loss = 0

    def step_update_lr(self):
        """it will update current iterations"""
        c1 = re.search(self.cfg.TRAIN.LR.SCHEDULER, 'Poly', re.IGNORECASE)
        c2 = re.search(self.cfg.TRAIN.LR.SCHEDULER, 'Cosine', re.IGNORECASE)
        if c1 or c2:
            self.scheduler.step()

        self.curr_step += 1

    def epoch_update_lr(self):
        c1 = re.search(self.cfg.TRAIN.LR.SCHEDULER, 'Poly', re.IGNORECASE)
        c2 = re.search(self.cfg.TRAIN.LR.SCHEDULER, 'Cosine', re.IGNORECASE)
        if not c1 and not c2:
            self.scheduler.step()

        self.curr_epoch += 1

    @staticmethod
    def get_major_prob(prob):
        if isinstance(prob, (list, tuple)):  # multi returns, such as deeply supervision
            prob = prob[0]
        return prob

    def get_optimizer_lr(self) -> float:
        """ Get optimizer learning rate """
        return self.optimizer.state_dict()['param_groups'][0]['lr']

    def tensor_to_gpu(self, im, gt):
        im = im.cuda()
        if self.cfg.MODEL.OUTPUT_CHANNELS > 1:  # using softmax, label dim is: B*H*W, sigmoid is: B*1*H*W
            gt = gt.squeeze().long()
        gt = gt.cuda()
        return im, gt

    def cal_batch_acc(self, prob, gt):
        """ cal batch accuracy for each batch during training and validate"""
        pred = utils.tensor_binarize(prob).flatten()  # get binary segmentation results
        mask = utils.tensor_to_numpy(gt).flatten()  # convert cuda tensor to numpy
        hist = utils.fast_hist(input=pred, label=mask, num_class=self.num_class)  # compute confusion matrix
        self.evaluator.confusion_matrix = hist
        _, acc_dict = self.evaluator.cal_acc(return_acc_dict=True)
        return hist, acc_dict

    @staticmethod
    def print_train_logs(log_head: str, loss: float, acc_dict: dict, phase='train'):
        logs = log_head + "loss: {:.6f}".format(loss)
        for key, value in acc_dict.items():
            logs = logs + " - " + key + ": " + str(value).rjust(5, ' ')
        logger = get_logger(phase)
        logger.info(logs)

    def report_epoch_logs(self, train_logs, valid_logs):
        key = self.cfg.TRAIN.KEY_METRIC
        TRAIN = self.cfg.TRAIN
        logger = get_logger('epoch: {}-{} training summary'.format(self.curr_epoch, self.cfg.TRAIN.EPOCHS))
        logger.critical("train - loss: {:.6f} - {}: {} - time: {} min".format(train_logs['loss'], key, train_logs[key], train_logs['time']))
        logger.critical("valid - loss: {:.6f} - {}: {} - time: {} min".format(valid_logs['loss'], key, valid_logs[key], valid_logs['time']))

        self.key_metric_record.append(float(valid_logs[TRAIN.KEY_METRIC]))
        j = _find_best_value_index(self.key_metric_record, TRAIN.HIGHER_BETTER)

        logger = get_logger('epoch: {}-{} training summary'.format(self.curr_epoch, TRAIN.EPOCHS))
        logger.critical('best {} record is at epoch - {}, best record - {:.2f}'.format(key, j + 1, self.key_metric_record[j]))

        t = train_logs['time'] + valid_logs['time']
        self.total_time = self.total_time + t
        remained_epoch = TRAIN.EPOCHS - self.curr_epoch
        remain_time = remained_epoch * t
        logger.warning('elapsed time: {:.2f} hours, {:.2f} hours remains for following training...'.format(
            self.total_time / 60,
            remain_time / 60))
        print('==' * 74)

    def save_checkpoint(self, model_states):
        ckpt = self.cfg.TRAIN.CKPT
        start_save_epoch = ckpt.START_SAVE_EPOCH
        save_path = os.path.join(ckpt.SAVE_DIR, 'ckpt.pth')

        if self.curr_epoch >= start_save_epoch:
            if not ckpt.DELETE_OLD:
                save_path = os.path.join(ckpt.SAVE_DIR, str(self.curr_epoch) + '_' + 'ckpt.pth')
                torch.save(model_states, save_path)
                self.curr_saved_epoch = self.curr_epoch
            else:
                if self.curr_epoch == start_save_epoch:
                    torch.save(model_states, save_path)
                    self.last_best_epoch = self.curr_epoch
                    self.curr_saved_epoch = self.curr_epoch
                else:
                    value = self.key_metric_record[start_save_epoch - 1:]
                    index = _find_best_value_index(value, self.cfg.TRAIN.HIGHER_BETTER)
                    curr_best_epoch = index + start_save_epoch

                    if curr_best_epoch >= self.curr_epoch:
                        os.remove(save_path)
                        torch.save(model_states, save_path)
                        self.last_best_epoch = curr_best_epoch
                        self.curr_saved_epoch = curr_best_epoch

        logger = get_logger('checkpoint saving info')
        print('==' * 74)
        if self.curr_saved_epoch is not None:
            logger.critical('currently saved checkpoint is at epoch - {}'.format(self.curr_saved_epoch))
        else:
            logger.critical('the process of saving model is not started')


class MixedPrecisionTrain(_BaseTrainer):
    def __init__(self, cfg, model, optimizer, scheduler, criterion, evaluator):
        super().__init__(cfg, model, optimizer, scheduler, criterion, evaluator)

    def forward_backward(self, im, gt):
        im, gt = self.tensor_to_gpu(im, gt)
        self.optimizer.zero_grad()  # zero model gradient

        with autocast():  # Runs the forward pass with autocasting.
            prob = self.model(im)
            loss = self.criterion(prob, gt)

            scaler.scale(loss).backward()  # gradient backward
            scaler.step(self.optimizer)  # optimizer parameters
            scaler.update()  # Updates the scale for next iteration.

        self.step_update_lr()  # update learning rate if scheduler is poly, cosine

        self.writer.add_scalar('train/step_lr', self.get_optimizer_lr(), self.curr_step)

        loss = loss.item()
        self.train_batch_loss += loss

        return prob, loss

    def forward(self, im, gt):
        """
        Return:
            prob: model prediction
            loss: loss value (not Tensor)
        """
        im, gt = self.tensor_to_gpu(im, gt)
        with autocast():  # Runs the forward pass with autocasting.
            prob = self.model(im)
            loss = self.criterion(prob, gt)

        loss = loss.item()
        self.valid_batch_loss += loss

        return prob, loss


class NormalTrain(_BaseTrainer):
    def __init__(self, cfg, model, optimizer, scheduler, criterion, evaluator):
        super().__init__(cfg, model, optimizer, scheduler, criterion, evaluator)

    def forward_backward(self, im, gt):
        im, gt = self.tensor_to_gpu(im, gt)

        self.optimizer.zero_grad()  # zero model gradient

        prob = self.model(im)
        loss = self.criterion(prob, gt)
        loss.backward()  # gradient backward
        self.optimizer.step()  # optimizer parameters

        self.step_update_lr()  # update learning rate if scheduler is poly, cosine

        loss = loss.item()
        self.train_batch_loss += loss

        return prob, loss

    def forward(self, im, gt):
        """
        Return:
            prob: model prediction
            loss: loss value (not Tensor)
        """
        im, gt = self.tensor_to_gpu(im, gt)
        prob = self.model(im)
        loss = self.criterion(prob, gt)

        loss = loss.item()
        self.valid_batch_loss += loss

        return prob, loss



# -*- encoding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021/4/22 下午8:02
Written by Yuwei Jin(yuwei_jin@163.com)
"""

import os
import logging

import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from torchvision.models.utils import load_state_dict_from_url

logger = logging.getLogger('hrnet_backbone')

__all__ = ['hrnet18', 'hrnet32', 'hrnet48']

model_urls = {
    'hrnet18': 'https://opr0mq.dm.files.1drv.com/y4mIoWpP2n-LUohHHANpC0jrOixm1FZgO2OsUtP2DwIozH5RsoYVyv_De5wDgR6XuQmirMV3C0AljLeB-zQXevfLlnQpcNeJlT9Q8LwNYDwh3TsECkMTWXCUn3vDGJWpCxQcQWKONr5VQWO1hLEKPeJbbSZ6tgbWwJHgHF7592HY7ilmGe39o5BhHz7P9QqMYLBts6V7QGoaKrr0PL3wvvR4w',
    'hrnet32': 'https://opr74a.dm.files.1drv.com/y4mKOuRSNGQQlp6wm_a9bF-UEQwp6a10xFCLhm4bqjDu6aSNW9yhDRM7qyx0vK0WTh42gEaniUVm3h7pg0H-W0yJff5qQtoAX7Zze4vOsqjoIthp-FW3nlfMD0-gcJi8IiVrMWqVOw2N3MbCud6uQQrTaEAvAdNjtjMpym1JghN-F060rSQKmgtq5R-wJe185IyW4-_c5_ItbhYpCyLxdqdEQ',
    'hrnet48': 'https://optgaw.dm.files.1drv.com/y4mWNpya38VArcDInoPaL7GfPMgcop92G6YRkabO1QTSWkCbo7djk8BFZ6LK_KHHIYE8wqeSAChU58NVFOZEvqFaoz392OgcyBrq_f8XGkusQep_oQsuQ7DPQCUrdLwyze_NlsyDGWot0L9agkQ-M_SfNr10ETlCF5R7BdKDZdupmcMXZc-IE3Ysw1bVHdOH4l-XEbEKFAi6ivPUbeqlYkRMQ'
}


# model_urls = {
#     'resnet18_ibn_a': 'https://github.com/XingangPan/IBN-Net/releases/download/v1.0/resnet18_ibn_a-2f571257.pth',
#     'resnet34_ibn_a': 'https://github.com/XingangPan/IBN-Net/releases/download/v1.0/resnet34_ibn_a-94bc1577.pth',
#     'resnet50_ibn_a': 'https://github.com/XingangPan/IBN-Net/releases/download/v1.0/resnet50_ibn_a-d9d0bb7b.pth',
#     'resnet101_ibn_a': 'https://github.com/XingangPan/IBN-Net/releases/download/v1.0/resnet101_ibn_a-59ea0ac6.pth',
#     'resnet18_ibn_b': 'https://github.com/XingangPan/IBN-Net/releases/download/v1.0/resnet18_ibn_b-bc2f3c11.pth',
#     'resnet34_ibn_b': 'https://github.com/XingangPan/IBN-Net/releases/download/v1.0/resnet34_ibn_b-04134c37.pth',
#     'resnet50_ibn_b': 'https://github.com/XingangPan/IBN-Net/releases/download/v1.0/resnet50_ibn_b-9ca61e85.pth',
#     'resnet101_ibn_b': 'https://github.com/XingangPan/IBN-Net/releases/download/v1.0/resnet101_ibn_b-c55f6dba.pth',
# }


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class HighResolutionModule(nn.Module):
    def __init__(self, num_branches, blocks, num_blocks, num_inchannels,
                 num_channels, fuse_method, multi_scale_output=True, norm_layer=None):
        super(HighResolutionModule, self).__init__()
        self._check_branches(
            num_branches, blocks, num_blocks, num_inchannels, num_channels)

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self.norm_layer = norm_layer

        self.num_inchannels = num_inchannels
        self.fuse_method = fuse_method
        self.num_branches = num_branches

        self.multi_scale_output = multi_scale_output

        self.branches = self._make_branches(
            num_branches, blocks, num_blocks, num_channels)
        self.fuse_layers = self._make_fuse_layers()
        self.relu = nn.ReLU(inplace=True)

    def _check_branches(self, num_branches, blocks, num_blocks,
                        num_inchannels, num_channels):
        if num_branches != len(num_blocks):
            error_msg = 'NUM_BRANCHES({}) <> NUM_BLOCKS({})'.format(
                num_branches, len(num_blocks))
            logger.error(error_msg)
            raise ValueError(error_msg)

        if num_branches != len(num_channels):
            error_msg = 'NUM_BRANCHES({}) <> NUM_CHANNELS({})'.format(
                num_branches, len(num_channels))
            logger.error(error_msg)
            raise ValueError(error_msg)

        if num_branches != len(num_inchannels):
            error_msg = 'NUM_BRANCHES({}) <> NUM_INCHANNELS({})'.format(
                num_branches, len(num_inchannels))
            logger.error(error_msg)
            raise ValueError(error_msg)

    def _make_one_branch(self, branch_index, block, num_blocks, num_channels,
                         stride=1):
        downsample = None
        if stride != 1 or \
                self.num_inchannels[branch_index] != num_channels[branch_index] * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.num_inchannels[branch_index],
                          num_channels[branch_index] * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                self.norm_layer(num_channels[branch_index] * block.expansion),
            )

        layers = []
        layers.append(block(self.num_inchannels[branch_index],
                            num_channels[branch_index], stride, downsample, norm_layer=self.norm_layer))
        self.num_inchannels[branch_index] = \
            num_channels[branch_index] * block.expansion
        for i in range(1, num_blocks[branch_index]):
            layers.append(block(self.num_inchannels[branch_index],
                                num_channels[branch_index], norm_layer=self.norm_layer))

        return nn.Sequential(*layers)

    def _make_branches(self, num_branches, block, num_blocks, num_channels):
        branches = []

        for i in range(num_branches):
            branches.append(
                self._make_one_branch(i, block, num_blocks, num_channels))

        return nn.ModuleList(branches)

    def _make_fuse_layers(self):
        if self.num_branches == 1:
            return None

        num_branches = self.num_branches
        num_inchannels = self.num_inchannels
        fuse_layers = []
        for i in range(num_branches if self.multi_scale_output else 1):
            fuse_layer = []
            for j in range(num_branches):
                if j > i:
                    fuse_layer.append(nn.Sequential(
                        nn.Conv2d(num_inchannels[j],
                                  num_inchannels[i],
                                  1,
                                  1,
                                  0,
                                  bias=False),
                        self.norm_layer(num_inchannels[i])))
                elif j == i:
                    fuse_layer.append(None)
                else:
                    conv3x3s = []
                    for k in range(i - j):
                        if k == i - j - 1:
                            num_outchannels_conv3x3 = num_inchannels[i]
                            conv3x3s.append(nn.Sequential(
                                nn.Conv2d(num_inchannels[j],
                                          num_outchannels_conv3x3,
                                          3, 2, 1, bias=False),
                                self.norm_layer(num_outchannels_conv3x3)))
                        else:
                            num_outchannels_conv3x3 = num_inchannels[j]
                            conv3x3s.append(nn.Sequential(
                                nn.Conv2d(num_inchannels[j],
                                          num_outchannels_conv3x3,
                                          3, 2, 1, bias=False),
                                self.norm_layer(num_outchannels_conv3x3),
                                nn.ReLU(inplace=True)))
                    fuse_layer.append(nn.Sequential(*conv3x3s))
            fuse_layers.append(nn.ModuleList(fuse_layer))

        return nn.ModuleList(fuse_layers)

    def get_num_inchannels(self):
        return self.num_inchannels

    def forward(self, x):
        if self.num_branches == 1:
            return [self.branches[0](x[0])]

        for i in range(self.num_branches):
            x[i] = self.branches[i](x[i])

        x_fuse = []
        for i in range(len(self.fuse_layers)):
            y = x[0] if i == 0 else self.fuse_layers[i][0](x[0])
            for j in range(1, self.num_branches):
                if i == j:
                    y = y + x[j]
                elif j > i:
                    width_output = x[i].shape[-1]
                    height_output = x[i].shape[-2]
                    y = y + F.interpolate(
                        self.fuse_layers[i][j](x[j]),
                        size=[height_output, width_output],
                        mode='bilinear',
                        align_corners=True
                    )
                else:
                    y = y + self.fuse_layers[i][j](x[j])
            x_fuse.append(self.relu(y))

        return x_fuse


blocks_dict = {
    'BASIC': BasicBlock,
    'BOTTLENECK': Bottleneck
}


class HighResolutionNet(nn.Module):

    def __init__(self,
                 cfg,
                 norm_layer=None):
        super(HighResolutionNet, self).__init__()

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self.norm_layer = norm_layer
        # stem network
        # stem net
        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1,
                               bias=False)
        self.bn1 = self.norm_layer(64)
        self.conv2 = nn.Conv2d(64, 64, kernel_size=3, stride=2, padding=1,
                               bias=False)
        self.bn2 = self.norm_layer(64)
        self.relu = nn.ReLU(inplace=True)

        # stage 1
        self.stage1_cfg = cfg['STAGE1']
        num_channels = self.stage1_cfg['NUM_CHANNELS'][0]
        block = blocks_dict[self.stage1_cfg['BLOCK']]
        num_blocks = self.stage1_cfg['NUM_BLOCKS'][0]
        self.layer1 = self._make_layer(block, 64, num_channels, num_blocks)
        stage1_out_channel = block.expansion * num_channels

        # stage 2
        self.stage2_cfg = cfg['STAGE2']
        num_channels = self.stage2_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage2_cfg['BLOCK']]
        num_channels = [
            num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition1 = self._make_transition_layer(
            [stage1_out_channel], num_channels)
        self.stage2, pre_stage_channels = self._make_stage(
            self.stage2_cfg, num_channels)

        # stage 3
        self.stage3_cfg = cfg['STAGE3']
        num_channels = self.stage3_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage3_cfg['BLOCK']]
        num_channels = [
            num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition2 = self._make_transition_layer(
            pre_stage_channels, num_channels)
        self.stage3, pre_stage_channels = self._make_stage(
            self.stage3_cfg, num_channels)

        # stage 4
        self.stage4_cfg = cfg['STAGE4']
        num_channels = self.stage4_cfg['NUM_CHANNELS']
        block = blocks_dict[self.stage4_cfg['BLOCK']]
        num_channels = [
            num_channels[i] * block.expansion for i in range(len(num_channels))]
        self.transition3 = self._make_transition_layer(
            pre_stage_channels, num_channels)
        self.stage4, pre_stage_channels = self._make_stage(
            self.stage4_cfg, num_channels, multi_scale_output=True)

        last_inp_channels = np.int(np.sum(pre_stage_channels))

        self.last_layer = nn.Sequential(
            nn.Conv2d(
                in_channels=last_inp_channels,
                out_channels=256,
                kernel_size=1,
                stride=1,
                padding=0),
            self.norm_layer(256),
            nn.ReLU(inplace=True)
        )

    def _make_transition_layer(
            self, num_channels_pre_layer, num_channels_cur_layer):
        num_branches_cur = len(num_channels_cur_layer)
        num_branches_pre = len(num_channels_pre_layer)

        transition_layers = []
        for i in range(num_branches_cur):
            if i < num_branches_pre:
                if num_channels_cur_layer[i] != num_channels_pre_layer[i]:
                    transition_layers.append(nn.Sequential(
                        nn.Conv2d(num_channels_pre_layer[i],
                                  num_channels_cur_layer[i],
                                  3,
                                  1,
                                  1,
                                  bias=False),
                        self.norm_layer(num_channels_cur_layer[i]),
                        nn.ReLU(inplace=True)))
                else:
                    transition_layers.append(None)
            else:
                conv3x3s = []
                for j in range(i + 1 - num_branches_pre):
                    inchannels = num_channels_pre_layer[-1]
                    outchannels = num_channels_cur_layer[i] \
                        if j == i - num_branches_pre else inchannels
                    conv3x3s.append(nn.Sequential(
                        nn.Conv2d(
                            inchannels, outchannels, 3, 2, 1, bias=False),
                        self.norm_layer(outchannels),
                        nn.ReLU(inplace=True)))
                transition_layers.append(nn.Sequential(*conv3x3s))

        return nn.ModuleList(transition_layers)

    def _make_layer(self, block, inplanes, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                self.norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(inplanes, planes, stride, downsample, norm_layer=self.norm_layer))
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(inplanes, planes, norm_layer=self.norm_layer))

        return nn.Sequential(*layers)

    def _make_stage(self, layer_config, num_inchannels,
                    multi_scale_output=True):
        num_modules = layer_config['NUM_MODULES']
        num_branches = layer_config['NUM_BRANCHES']
        num_blocks = layer_config['NUM_BLOCKS']
        num_channels = layer_config['NUM_CHANNELS']
        block = blocks_dict[layer_config['BLOCK']]
        fuse_method = layer_config['FUSE_METHOD']

        modules = []
        for i in range(num_modules):
            # multi_scale_output is only used last module
            if not multi_scale_output and i == num_modules - 1:
                reset_multi_scale_output = False
            else:
                reset_multi_scale_output = True

            modules.append(
                HighResolutionModule(num_branches,
                                     block,
                                     num_blocks,
                                     num_inchannels,
                                     num_channels,
                                     fuse_method,
                                     reset_multi_scale_output,
                                     norm_layer=self.norm_layer)
            )
            num_inchannels = modules[-1].get_num_inchannels()

        return nn.Sequential(*modules), num_inchannels

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.layer1(x)

        x_list = []
        for i in range(self.stage2_cfg['NUM_BRANCHES']):
            if self.transition1[i] is not None:
                x_list.append(self.transition1[i](x))
            else:
                x_list.append(x)
        y_list = self.stage2(x_list)

        x_list = []
        for i in range(self.stage3_cfg['NUM_BRANCHES']):
            if self.transition2[i] is not None:
                if i < self.stage2_cfg['NUM_BRANCHES']:
                    x_list.append(self.transition2[i](y_list[i]))
                else:
                    x_list.append(self.transition2[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        y_list = self.stage3(x_list)

        x_list = []
        for i in range(self.stage4_cfg['NUM_BRANCHES']):
            if self.transition3[i] is not None:
                if i < self.stage3_cfg['NUM_BRANCHES']:
                    x_list.append(self.transition3[i](y_list[i]))
                else:
                    x_list.append(self.transition3[i](y_list[-1]))
            else:
                x_list.append(y_list[i])
        x = self.stage4(x_list)

        # Upsampling
        x0_h, x0_w = x[0].size(2), x[0].size(3)
        # x1 = F.interpolate(x[1], size=(x0_h, x0_w), mode='bilinear', align_corners=True)
        # x2 = F.interpolate(x[2], size=(x0_h, x0_w), mode='bilinear', align_corners=True)
        # x3 = F.interpolate(x[3], size=(x0_h, x0_w), mode='bilinear', align_corners=True)
        #
        # x = torch.cat([x[0], x1, x2, x3], 1)
        #
        # x = self.last_layer(x)

        return x


def _hrnet(arch, pretrained, **kwargs):
    try:
        from ..config.hrnet_config import MODEL_CONFIGS
    except ImportError:
        from sv.libs import MODEL_CONFIGS
    model = HighResolutionNet(MODEL_CONFIGS[arch], **kwargs)
    if pretrained:
        ckpt_folder = os.path.split(os.path.realpath(__file__))[0]
        model_urls = {'hrnet18': os.path.join(ckpt_folder, 'hrnetv2_w18_imagenet_pretrained.pth'),
                      'hrnet32': os.path.join(ckpt_folder, 'hrnetv2_w32_imagenet_pretrained.pth'),
                      'hrnet40': os.path.join(ckpt_folder, 'hrnetv2_w40_imagenet_pretrained.pth'),
                      'hrnet48': os.path.join(ckpt_folder, 'hrnetv2_w48_imagenet_pretrained.pth')
                      }

        state_dict = torch.load(model_urls[arch])
        model.load_state_dict(state_dict, strict=False)
    return model


def hrnet18(pretrained=True, **kwargs):
    r"""HRNet-18 model
    """
    return _hrnet('hrnet18', pretrained, **kwargs)


def hrnet32(pretrained=True, **kwargs):
    r"""HRNet-32 model
    """
    return _hrnet('hrnet32', pretrained, **kwargs)


def hrnet40(pretrained=True, **kwargs):
    r"""HRNet-40 model
    """
    return _hrnet('hrnet40', pretrained, **kwargs)


def hrnet48(pretrained=True, **kwargs):
    r"""HRNet-48 model
    """
    return _hrnet('hrnet48', pretrained, **kwargs)


if __name__ == '__main__':
    model = hrnet40(pretrained=True)
    x = torch.randn(1, 3, 512, 512)
    model(x)



# -*- coding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2020-12-19 21:39
Written by Yuwei Jin (642281525@qq.com)
"""
import math
import warnings

import torch

from sv.libs.data.dataset.dataset import load_train_data, load_val_data
from sv.libs.core.trainer import MixedPrecisionTrain
from sv.libs.tools import utils


class _Train(MixedPrecisionTrain):
    def __init__(self, cfg, model, optimizer, criterion, evaluator, scheduler):

        model = model.cuda()
        warnings.warn('Note, only single GPU device is supported!')

        self.train_data = load_train_data(cfg)
        self.valid_data = load_val_data(cfg)

        # parse learning rate scheduler
        self.scheduler = utils.parse_lr_scheduler(cfg, optimizer, train_data=self.train_data)
        warnings.warn('Note, only Cosine, Poly, and ExponentialLR scheduler are supported!')

        # some hyper-params
        self.num_class = cfg.DATASET.NUM_CLASS
        self.epochs = cfg.TRAIN.EPOCHS
        self.train_batch_steps = math.floor(len(self.train_data.dataset) / cfg.TRAIN.BATCH_SIZE)
        self.valid_batch_steps = math.floor(len(self.valid_data.dataset) / cfg.TRAIN.BATCH_SIZE)

        self.cfg = cfg

        super().__init__(cfg, model, optimizer, scheduler, criterion, evaluator)

    def _show_logs(self, loss, prob, gt, step, batch_steps, phase='train'):
        conf_mat, acc_dict = self.cal_batch_acc(prob, gt)  # compute confusion matrix, acc, acc dict
        if step % self.cfg.TRAIN.MSG_ITER == 0:
            log_head = "epoch: {}-{} - iter: {}-{} | ".format(self.curr_epoch, self.epochs, step + 1, batch_steps)
            self.print_train_logs(log_head, loss, acc_dict, phase=phase)

        return conf_mat

    def _epoch_train(self):
        self.model.train()
        for i, (im, gt) in enumerate(self.train_data):
            prob, loss = self.forward_backward(im, gt)
            prob = self.get_major_prob(prob)  # multi returns, such as deeply supervision
            conf_mat = self._show_logs(loss, prob, gt, i, self.train_batch_steps)
            self.train_batch_hist += conf_mat

    def _epoch_eval(self):
        self.model.eval()
        with torch.no_grad():
            for i, (im, gt) in enumerate(self.valid_data):
                prob, loss = self.forward(im, gt)
                prob = self.get_major_prob(prob)
                conf_mat = self._show_logs(loss, prob, gt, i, self.valid_batch_steps, 'valid')
                self.valid_batch_hist += conf_mat

    def train(self):
        print("---start training...")
        for epoch in range(self.epochs):
            time_counter = utils.TimeCounter()

            time_counter.tic()
            self._epoch_train()  # call train
            train_time = time_counter.toc()

            time_counter.tic()
            self._epoch_eval()  # call eval

            # calculate epoch accuracy
            self.evaluator.confusion_matrix = self.train_batch_hist
            train_acc, train_logs = self.evaluator.cal_acc(return_acc_dict=True)
            self.evaluator.confusion_matrix = self.valid_batch_hist
            valid_acc, valid_logs = self.evaluator.cal_acc(return_acc_dict=True)

            model_states = {"net": self.model.state_dict(), "optimizer": self.optimizer.state_dict(), "epoch": epoch + 1}
            self.save_checkpoint(model_states)
            valid_time = time_counter.toc()

            # calculate epoch average loss
            train_logs['loss'] = self.train_batch_loss / self.train_batch_steps
            valid_logs['loss'] = self.valid_batch_loss / self.valid_batch_steps
            train_logs['time'] = round(train_time, 2)
            valid_logs['time'] = round(valid_time, 2)

            self.writer.add_scalar('train/epoch_lr', self.get_optimizer_lr(), epoch)
            self.writer.add_scalars('train/time', {'train': train_time, 'val': valid_time}, epoch)
            for key in self.employed_metrics:
                self.writer.add_scalars('train/' + key, {'train': float(train_logs[key]), 'val': float(valid_logs[key])}, epoch)

            self.report_epoch_logs(train_logs, valid_logs)

            self.epoch_update_lr()
            self.epoch_reset()

        self.writer.close()

        print('-------------Congratulations! Training Done!!!-------------')

if __name__ == '__main__':
    from nb_log import LogManager


    class NbLog:
        def __init__(self):
            self.logger = LogManager('simple').get_logger_and_add_handlers(log_path="log 存放地址", log_filename="日志名称")

        def error(self, msg):
            return self.logger.error(msg)

        def debug(self, msg):
            return self.logger.debug(msg)

        def info(self, msg):
            return self.logger.info(msg)

        def warning(self, msg):
            return self.logger.warning(msg)

        def exception(self, msg, exc_info=True):
            return self.logger.exception(msg, exc_info)


    dolog = NbLog()
    dolog.debug("这个是 debug 日志")
    dolog.error("这个是 error 日志")


# -*- encoding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021/4/22 下午5:01
Written by Yuwei Jin(yuwei_jin@163.com)
"""

import torch.nn as nn
import torch

from torchvision import models


class VGGNet(nn.Module):
    def __init__(self, pretrained=True):
        super(VGGNet, self).__init__()

        self.vgg = models.vgg16_bn(pretrained=pretrained)

        del self.vgg.classifier

    def forward(self, x):
        layer1 = self.vgg.features[0:6](x)  # 64*512*512
        layer2 = self.vgg.features[6:13](layer1)  # 128*256*256
        layer3 = self.vgg.features[13:23](layer2)  # 256*128*128
        layer4 = self.vgg.features[23:33](layer3)  # 512*64*64
        layer5 = self.vgg.features[33:43](layer4)  # 512*32*32
        return layer1, layer2, layer3, layer4, layer5


if __name__ == '__main__':
    x = torch.randn((1, 3, 512, 512))
    model = VGGNet().eval()
    layer1, layer2, layer3, layer4, layer5 = model(x)
    print(layer5.size())

    print(models.vgg16_bn(pretrained=True))



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-23 19:16 
Written by Yuwei Jin (642281525@qq.com)
"""
import torch
from torch import nn
from torch.nn import functional as F

from sv.libs.tools.init_weight import init_weights
from sv.modules.packaged_conv import Conv3x3_BN_ReLU, Conv1x1_BN_ReLU


class AtrousPyramidPooling(nn.Module):
    """ASPP: ref DeepLab v3+"""
    def __init__(self, in_chs, reduction_dim=256, rates=(1, 6, 12, 18)):
        super(AtrousPyramidPooling, self).__init__()

        self.pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            Conv1x1_BN_ReLU(in_chs, reduction_dim)
        )

        self.conv1 = Conv3x3_BN_ReLU(in_chs, reduction_dim, dilation=rates[0])
        self.conv2 = Conv3x3_BN_ReLU(in_chs, reduction_dim, dilation=rates[1])
        self.conv3 = Conv3x3_BN_ReLU(in_chs, reduction_dim, dilation=rates[2])
        self.conv4 = Conv3x3_BN_ReLU(in_chs, reduction_dim, dilation=rates[3])

        self.bottle_conv = nn.Sequential(
            Conv1x1_BN_ReLU(5 * reduction_dim, reduction_dim),
            nn.Dropout2d(p=0.2)
        )

    def forward(self, x):
        pool = self.pool(x)
        pool = F.upsample(input=pool, size=x.size()[2:], mode='bilinear', align_corners=True)

        x1 = self.conv1(x)
        x2 = self.conv2(x)
        x3 = self.conv3(x)
        x4 = self.conv4(x)

        x = torch.cat([pool, x1, x2, x3, x4], dim=1)

        return self.bottle_conv(x)


class PPM(nn.Module):
    """Pyramid Pooling Module: ref PSPNet"""
    def __init__(self, in_chs, reduction_dim, out_chs, bins=(1, 2, 3, 6)):
        super().__init__()
        self.stages = nn.ModuleList([self._make_stage(in_chs, reduction_dim, b) for b in bins])

        self.bottle_conv = nn.Sequential(
            nn.Conv2d(in_chs * 2, out_chs, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_chs),
            nn.ReLU(inplace=True),
            nn.Dropout2d(p=0.2)
        )

        init_weights(self.stages, self.bottle_conv)

    @staticmethod
    def _make_stage(in_chs, reduction_dim, bin):
        return nn.Sequential(
            nn.AdaptiveAvgPool2d(bin),
            nn.Conv2d(in_chs, reduction_dim, kernel_size=1, bias=False),
            nn.BatchNorm2d(reduction_dim),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        H, W = x.size()[2:]

        out = x
        for f in self.stages:
            y = f(x)
            y = F.interpolate(input=y, size=(H, W), mode='bilinear', align_corners=True)
            out = torch.cat([out, y], dim=1)

        return self.bottle_conv(out)


class FPA(nn.Module):
    def __init__(self, channels=2048):
        """
        Feature Pyramid Attention
        :type channels: int
        """
        super(FPA, self).__init__()
        channels_mid = int(channels/4)

        self.channels_cond = channels

        # Master branch
        self.conv_master = nn.Conv2d(self.channels_cond, channels, kernel_size=1, bias=False)
        self.bn_master = nn.BatchNorm2d(channels)

        # Global pooling branch
        self.conv_gpb = nn.Conv2d(self.channels_cond, channels, kernel_size=1, bias=False)
        self.bn_gpb = nn.BatchNorm2d(channels)

        # C333 because of the shape of last feature maps is (16, 16).
        self.conv7x7_1 = nn.Conv2d(self.channels_cond, channels_mid, kernel_size=(7, 7), stride=2, padding=3, bias=False)
        self.bn1_1 = nn.BatchNorm2d(channels_mid)
        self.conv5x5_1 = nn.Conv2d(channels_mid, channels_mid, kernel_size=(5, 5), stride=2, padding=2, bias=False)
        self.bn2_1 = nn.BatchNorm2d(channels_mid)
        self.conv3x3_1 = nn.Conv2d(channels_mid, channels_mid, kernel_size=(3, 3), stride=2, padding=1, bias=False)
        self.bn3_1 = nn.BatchNorm2d(channels_mid)

        self.conv7x7_2 = nn.Conv2d(channels_mid, channels_mid, kernel_size=(7, 7), stride=1, padding=3, bias=False)
        self.bn1_2 = nn.BatchNorm2d(channels_mid)
        self.conv5x5_2 = nn.Conv2d(channels_mid, channels_mid, kernel_size=(5, 5), stride=1, padding=2, bias=False)
        self.bn2_2 = nn.BatchNorm2d(channels_mid)
        self.conv3x3_2 = nn.Conv2d(channels_mid, channels_mid, kernel_size=(3, 3), stride=1, padding=1, bias=False)
        self.bn3_2 = nn.BatchNorm2d(channels_mid)

        # Convolution Upsample
        self.conv_upsample_3 = nn.ConvTranspose2d(channels_mid, channels_mid, kernel_size=4, stride=2, padding=1, bias=False)
        self.bn_upsample_3 = nn.BatchNorm2d(channels_mid)

        self.conv_upsample_2 = nn.ConvTranspose2d(channels_mid, channels_mid, kernel_size=4, stride=2, padding=1, bias=False)
        self.bn_upsample_2 = nn.BatchNorm2d(channels_mid)

        self.conv_upsample_1 = nn.ConvTranspose2d(channels_mid, channels, kernel_size=4, stride=2, padding=1, bias=False)
        self.bn_upsample_1 = nn.BatchNorm2d(channels)

        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        """
        :param x: Shape: [b, 2048, h, w]
        :return: out: Feature maps. Shape: [b, 2048, h, w]
        """
        # Master branch
        x_master = self.conv_master(x)
        x_master = self.bn_master(x_master)

        # Global pooling branch
        x_gpb = nn.AvgPool2d(x.shape[2:])(x).view(x.shape[0], self.channels_cond, 1, 1)
        x_gpb = self.conv_gpb(x_gpb)
        x_gpb = self.bn_gpb(x_gpb)

        # Branch 1
        x1_1 = self.conv7x7_1(x)
        x1_1 = self.bn1_1(x1_1)
        x1_1 = self.relu(x1_1)
        x1_2 = self.conv7x7_2(x1_1)
        x1_2 = self.bn1_2(x1_2)

        # Branch 2
        x2_1 = self.conv5x5_1(x1_1)
        x2_1 = self.bn2_1(x2_1)
        x2_1 = self.relu(x2_1)
        x2_2 = self.conv5x5_2(x2_1)
        x2_2 = self.bn2_2(x2_2)

        # Branch 3
        x3_1 = self.conv3x3_1(x2_1)
        x3_1 = self.bn3_1(x3_1)
        x3_1 = self.relu(x3_1)
        x3_2 = self.conv3x3_2(x3_1)
        x3_2 = self.bn3_2(x3_2)

        # Merge branch 1 and 2
        x3_upsample = self.relu(self.bn_upsample_3(self.conv_upsample_3(x3_2)))
        x2_merge = self.relu(x2_2 + x3_upsample)
        x2_upsample = self.relu(self.bn_upsample_2(self.conv_upsample_2(x2_merge)))
        x1_merge = self.relu(x1_2 + x2_upsample)

        x_master = x_master * self.relu(self.bn_upsample_1(self.conv_upsample_1(x1_merge)))

        #
        out = self.relu(x_master + x_gpb)

        return out



# -*- encoding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021/5/14 上午8:53
Written by Yuwei Jin (yuwei_jin@163.com)
"""

import os
import cv2
import math
import json

import numpy as np
from PIL import Image
from multiprocessing import Pool

from sv.libs.tools import utils


class DilatedCrop:
    """Crop a big image and corresponding label into small patches
        mainly used for inferring.
    """

    def __init__(self, crop_cfg, use_pil_read=False):
        self.patch_size = crop_cfg['patch_size']
        self.win = crop_cfg['win_size']
        self.im_suffix = crop_cfg['im_suffix']
        self.gt_suffix = crop_cfg['gt_suffix']
        self.dataset_cfg = crop_cfg['dataset_cfg']

        self.dataset_name = crop_cfg['dataset_name']
        root_dir = crop_cfg['root_dir']

        output_root_dir = crop_cfg['output_root_dir']
        output_root_dir = os.path.join(output_root_dir, 'test')

        self.im_src = os.path.join(root_dir, 'image')
        self.gt_src = os.path.join(root_dir, 'label')

        self.im_output_src = os.path.join(output_root_dir, 'image')
        self.gt_output_src = os.path.join(output_root_dir, 'label')
        self.crop_info_dir = os.path.join(output_root_dir, 'crop_info')

        self.use_pil_read = use_pil_read


    def save_crop_info(self, im, im_name):
        rows, cols = im.shape[0], im.shape[1]
        m = math.ceil(rows / self.win)
        n = math.ceil(cols / self.win)

        info = {'rows': rows,
                'cols': cols,
                'patch_size': self.patch_size,
                'central_win_size': self.win,
                'm': m,
                'n': n
                }

        with open(os.path.join(self.crop_info_dir, im_name + '_crop_info.json'), 'w') as f:
            f.write(json.dumps(info))

    def _crop(self, im_name):
        """im_name example: test1.tif"""
        # applying PIL to load big scale images
        if self.use_pil_read:
            Image.MAX_IMAGE_PIXELS = None
            im = Image.open(os.path.join(self.im_src, im_name)).convert('RGB')
            gt = Image.open(os.path.join(self.gt_src, im_name)).convert('L')
            im = np.array(im)
            gt = np.array(gt)
        else:
            im = cv2.imread(os.path.join(self.im_src, im_name))
            gt = cv2.imread(os.path.join(self.gt_src, im_name), cv2.IMREAD_GRAYSCALE)

        print('processing: ' + im_name)
        im_name, _ = im_name.split('.')
        self.save_crop_info(im, im_name)

        # pad image and label
        rows, cols = im.shape[0], im.shape[1]
        win = self.win
        patch_size = self.patch_size
        m = math.ceil(rows / win)
        n = math.ceil(cols / win)
        pad = (patch_size - win) // 2

        top = pad
        bottom = win * m - rows + pad
        left = pad
        right = win * n - cols + pad

        mean, std = utils.load_mean_std(self.dataset_cfg)
        im = cv2.copyMakeBorder(im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=mean)
        gt = cv2.copyMakeBorder(gt, top, bottom, left, right, cv2.BORDER_CONSTANT, value=0)

        if self.use_pil_read:
            im = cv2.cvtColor(im, cv2.COLOR_RGB2BGR)

        k = 0
        for i in range(m):
            x1 = i * win + top - pad
            x2 = x1 + win + 2 * pad
            if i == 0:
                x1, x2 = 0, patch_size

            for j in range(n):
                y1 = j * win - pad + left
                y2 = y1 + win + 2 * pad
                if j == 0:
                    y1, y2 = 0, patch_size

                im_patch = im[x1:x2, y1:y2, :]
                gt_patch = gt[x1:x2, y1:y2]

                k += 1
                im = cv2.cvtColor(im, cv2.COLOR_RGB2BGR)
                cv2.imwrite(os.path.join(self.im_output_src, im_name + '-' + str(k) + self.im_suffix), im_patch)
                cv2.imwrite(os.path.join(self.gt_output_src, im_name + '-' + str(k) + self.im_suffix), gt_patch)

    @utils.time_counter
    def crop(self):
        utils.mk_dirs_r(self.im_output_src)
        utils.mk_dirs_r(self.gt_output_src)
        utils.mk_dirs_r(self.crop_info_dir)

        im_lists = utils.get_im_name_list(self.im_src, self.im_suffix)
        if len(im_lists) >= 8:  # applying multi processing
            pool = Pool(processes=16)
            pool.map(self._crop, im_lists)
            pool.close()
            pool.join()
        else:
            for f in im_lists:
                self._crop(f)



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-05-04 15:28
Written by Yuwei Jin (642281525@qq.com)
"""

import torch.nn as nn


def _real_init_weights(m, nonlinearity='relu'):
    if isinstance(m, list):
        for mini_m in m:
            _real_init_weights(mini_m)
    else:
        if isinstance(m, nn.Conv2d):
            # print('initializing Convolution layer...')
            nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity=nonlinearity)
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm, nn.InstanceNorm2d)):
            # print('initializing BN layer...')
            nn.init.ones_(m.weight)
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Linear):
            # print('initializing Linear layer...')
            nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity=nonlinearity)
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Module):
            for mini_m in m.children():
                _real_init_weights(mini_m)
        else:
            raise NotImplementedError('Unknown init module.')


def init_weights(*models, nonlinearity='relu'):
    """
    Initialize model's modules.
    """
    print('initializing modules...')
    for model in models:
        _real_init_weights(model, nonlinearity=nonlinearity)


# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-19 14:38
Written by Yuwei Jin (642281525@qq.com)
"""
import os
import numpy as np
import torch
from abc import ABC, abstractmethod
from torch.cuda.amp import autocast
from torchvision.transforms.functional import vflip, hflip
from torch.nn import functional as F


def resize_tensor(input, size, resize_mode='bilinear'):
    if not isinstance(input, torch.Tensor):
        raise ValueError('Input should be torch.Tensor type. Got unexpected type:{}'.format(type(input)))

    if resize_mode == 'bilinear':
        out = F.interpolate(input, size=size, mode=resize_mode, align_corners=True)
    elif resize_mode == 'nearset':
        out = F.interpolate(input, size=size, mode=resize_mode)
    else:
        raise NotImplementedError()

    return out


class Infer(ABC):
    def __init__(self, model, cfg):
        model = model().cuda()
        try:
            ckpt_path = os.path.join(cfg.TRAIN.CKPT.SAVE_DIR, 'ckpt.pth')
            weight = torch.load(ckpt_path)['net']
            model.load_state_dict(weight)
        except FileNotFoundError:
            raise FileNotFoundError('saved model weight is not found in dir: {}'.format(cfg.TRAIN.CKPT.SAVE_DIR))
        self.model = model.eval()
        num_class = cfg.DATASET.NUM_CLASS
        self.conf_mat = np.zeros((num_class, num_class))
        self.cfg = cfg

    def _infer(self, im: torch.Tensor) -> torch.Tensor:
        im = im.cuda()
        if self.cfg.TEST.H_FLIP and self.cfg.TEST.V_FLIP:
            x = torch.cat([im, hflip(im)], dim=0)
            x = torch.cat([x, vflip(im)], dim=0)
            y = self.model(x)
            out = y[0, :, :, :] + hflip(y[1, :, :, :]) + vflip(y[2, :, :, :])
            return out.unsqueeze(dim=0) / 3.0
        elif self.cfg.TEST.H_FLIP and not self.cfg.TEST.V_FLIP:
            x = torch.cat([im, hflip(im)], dim=0)
            y = self.model(x)
            out = y[0, :, :, :] + hflip(y[1, :, :, :])
            return out.unsqueeze(dim=0) / 2.0
        elif not self.cfg.TEST.H_FLIP and self.cfg.TEST.V_FLIP:
            x = torch.cat([im, vflip(im)], dim=0)
            y = self.model(x)
            out = y[0, :, :, :] + vflip(y[1, :, :, :])
            return out.unsqueeze(dim=0) / 2.0
        else:
            return self.model(im)

    def _multi_scale_infer(self, im: torch.Tensor) -> torch.Tensor:
        """ multi scale inference """
        _, _, ori_h, ori_w = im.size()
        final_prob = torch.zeros([1, self.cfg.MODEL.OUTPUT_CHANNELS, ori_h, ori_w]).cuda()

        for scale in self.cfg.TEST.MS_SCALE_LIST:
            new_h = int(ori_h * scale)
            new_w = int(ori_w * scale)
            new_im = resize_tensor(input=im, size=[new_h, new_w])

            prob = self._infer(new_im)
            prob = resize_tensor(prob, (ori_h, ori_w))
            final_prob += prob

        final_prob /= len(self.cfg.TEST.MS_SCALE_LIST)

        return final_prob

    def forward(self, x):
        if self.cfg.TEST.MS:
            out = self._multi_scale_infer(x)
        else:
            out = self._infer(x)
        return out



# -*- coding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2020-11-10 14:39
Written by Yuwei Jin (642281525@qq.com)
"""

import os
import multiprocessing
from yacs.config import CfgNode as CN

# TODO: It is necessary to change the following three items according to the specific model and data

model_name = 'res18_fpn'
dataset_name = 'WHU_Overlapped'  # Vaihingen
dilated_predict = True
seed = 1024  #

# -----------------------------------------------------------------------------------------------------
_C = CN()

# Dataset related settings
_C.DATASET = CN()
_C.DATASET.NAME = dataset_name  # Massachusetts WHU
_C.DATASET.ROOT = '/home/jyw/My_DatasetPatches/BE'  # dataset root dir
_C.DATASET.MEAN_STD = 'imagenet'
_C.DATASET.TRAIN_SET = os.path.join(_C.DATASET.ROOT, _C.DATASET.NAME, 'train')
_C.DATASET.VALID_SET = os.path.join(_C.DATASET.ROOT, _C.DATASET.NAME, 'val')
_C.DATASET.TEST_SET = os.path.join(_C.DATASET.ROOT, _C.DATASET.NAME, 'test')
_C.DATASET.TEST_WHOLE_LABEL = os.path.join(_C.DATASET.ROOT, _C.DATASET.NAME, 'whole_test_img_label', 'image')
_C.DATASET.TEST_WHOLE_IMAGE = os.path.join(_C.DATASET.ROOT, _C.DATASET.NAME, 'whole_test_img_label', 'label')
_C.DATASET.IMG_EXTENSION = '.tif'
_C.DATASET.NUM_CLASS = 2
_C.DATASET.IGNORE_LABEL = 0

# Common params for NETWORK
_C.MODEL = CN()
_C.MODEL.NAME = model_name
_C.MODEL.OUTPUT_STRIDE = 8  # for resnet series to control the output stride, optional[8,16,32]
_C.MODEL.BACKBONE = 'resnet50'
_C.MODEL.USE_STEM = True
_C.MODEL.INTERPOLATE_ALIGN = True
_C.MODEL.OUTPUT_DIR = os.path.join('results', _C.MODEL.NAME + '_' + _C.DATASET.NAME)
_C.MODEL.OUTPUT_CHANNELS = 1

# Training related params
_C.TRAIN = CN()
_C.TRAIN.SEED = seed
_C.TRAIN.CUDNN_DETERMINISTIC = False
_C.TRAIN.EPOCHS = 100
_C.TRAIN.BATCH_SIZE = 6
_C.TRAIN.INIT_LR = 0.001
_C.TRAIN.OPTIMIZER = 'adam'
_C.TRAIN.MOMENTUM = 0.9
_C.TRAIN.WD = 5e-4  # Weight decay
_C.TRAIN.FP16 = False
_C.TRAIN.MSG_ITER = 10
_C.TRAIN.RESUME = False  # 加载模型基础上继续训练
_C.TRAIN.LOGS_SAVE_DIR = os.path.join(_C.MODEL.OUTPUT_DIR, 'train_logs')
# support
_C.TRAIN.KEY_METRIC = 'iou'
_C.TRAIN.HIGHER_BETTER = True
_C.TRAIN.EMPLOYED_METRICS = ['pre', 'rec', 'f1', 'iou', 'miou']

_C.TRAIN.LR = CN()
_C.TRAIN.LR.SCHEDULER = 'Poly'
_C.TRAIN.LR.POWER = 0.9
_C.TRAIN.LR.WARMUP_START_LR = 5e-1
_C.TRAIN.LR.STEP_SIZE = 20
_C.TRAIN.LR.WARMUP_EPOCH = 5  # period of warmup
_C.TRAIN.LR.MILESTONES = None  # milestones for multi-step scheduler
_C.TRAIN.LR.ADJUST_PERIOD = None  # period of adjust learning rate
_C.TRAIN.LR.ETA_MIN_LR = 0.0  # minimum learning rate
_C.TRAIN.LR.T_MAX = _C.TRAIN.EPOCHS  # for CosineAnnealing decay

# checkpoint saving settings
_C.TRAIN.CKPT = CN()
_C.TRAIN.CKPT.SAVE_DIR = os.path.join(_C.MODEL.OUTPUT_DIR, 'checkpoint')
_C.TRAIN.CKPT.START_SAVE_EPOCH = 1
_C.TRAIN.CKPT.DELETE_OLD = False
_C.TRAIN.CKPT.SAVE_DIR = os.path.join(_C.MODEL.OUTPUT_DIR, 'checkpoint')

# data augmentation settings
_C.TRAIN.AUG = CN()
_C.TRAIN.AUG.MULTI_SCALE = True
_C.TRAIN.AUG.CENTER_CROP = False
# TODO: if the size of input data has changed, it must be changed.
_C.TRAIN.AUG.CROP_SIZE = [512, 512]
_C.TRAIN.AUG.SCALE_LIST = [0.75, 1.0, 1.25, 1.5, 1.75]
_C.TRAIN.AUG.MS_P = 0.5
# rotation
_C.TRAIN.AUG.ROTATION = False
_C.TRAIN.AUG.ROTATION_ANGLE = [90, 180, 270]
_C.TRAIN.AUG.ROTATION_P = 0.5
# flip
_C.TRAIN.AUG.FLIP = False
_C.TRAIN.AUG.FLIP_P = 0.5
# color jitter
_C.TRAIN.AUG.COLOR_JITTER = False
_C.TRAIN.AUG.C0LOR_JITTER_P = 0.5
# How much to adjust the brightness. Can be any non negative number.
# 0 gives a black image, 1 gives the original image while 2 increases the brightness by a factor of 2.
_C.TRAIN.AUG.BRIGHTNESS_FACTOR = 0.5
# How much to adjust the saturation.
# 0 will give a black and white image, 1 will give the original image while 2 will enhance the saturation by a factor of 2.
_C.TRAIN.AUG.SATURATION_FACTOR = 0.5
# How much to adjust the contrast. Can be any non negative number.
# 0 gives a solid gray image, 1 gives the original image while 2 increases the contrast by a factor of 2.
_C.TRAIN.AUG.CONTRAST_FACTOR = 0.5
# is the amount of shift in H channel and must be in the
#     interval `[-0.5, 0.5]`.
_C.TRAIN.AUG.HUE_FACTOR = 0.5

_C.TRAIN.AUG.GAUSSIAN_BLUR = False
_C.TRAIN.AUG.GAUSSIAN_BLUR_RADIUS = 2
_C.TRAIN.AUG.GAUSSIAN_BLUR_P = 0.5

_C.TRAIN.AUG.GAUSSIAN_NOISE = False
_C.TRAIN.AUG.GAUSSIAN_NOISE_MEAN = 0
_C.TRAIN.AUG.GAUSSIAN_NOISE_VAR = 0.001
_C.TRAIN.AUG.GAUSSIAN_NOISE_P = 0.5

# dataloader related params
_C.DATALOADER = CN()
_C.DATALOADER.WORKERS = 4
_C.DATALOADER.PIP_MEMORY = True
_C.DATALOADER.DROP_LAST = True

# test related params
_C.TEST = CN()
_C.TEST.NORMAL = False
_C.TEST.V_FLIP = False
_C.TEST.H_FLIP = False
_C.TEST.MS = False
_C.TEST.MS_SCALE_LIST = [0.75, 1, 1.25, 1.5, 1.75]
_C.TEST.FP16_VAL = False
_C.TEST.METRICS = ['pre', 'rec', 'f1', 'iou', 'miou']

_C.TEST.USE_DILATED_PREDICT = dilated_predict

# saving settings for test
_C.TEST.OUTPUT_DIR = os.path.join(_C.MODEL.OUTPUT_DIR, 'test_results')
_C.TEST.SAVE_RES = True
_C.TEST.PROB_SAVE_DIR = os.path.join(_C.TEST.OUTPUT_DIR, 'cropped_im_res')
_C.TEST.STITCH_RES = True  # some predicted results of the cropped images may be needed to stitch.
_C.TEST.STITCH_SAVE_RES = True
_C.TEST.STITCH_SAVE_DIR = os.path.join(_C.TEST.OUTPUT_DIR, 'whole_im_res')

# result palette
_C.TEST.BACKGROUND_COLOR = [0, 0, 0]
_C.TEST.TP_COLOR = [0, 0, 255]
_C.TEST.FP_COLOR = [0, 255, 255]
_C.TEST.FN_COLOR = [0, 255, 30]
_C.TEST.ALPHA = 0.5


def update_config():
    cfg = _C.clone()
    if dataset_name is not None:
        this_dir = os.path.dirname(__file__)
        cfg_file = os.path.join(this_dir, dataset_name + '.yaml')
        cfg.merge_from_file(cfg_file)
        cfg.freeze()

    return cfg


cfg = update_config()

if __name__ == '__main__':
    print(cfg.TEST.USE_DILATED_PREDICT)

    import json

    data = json.dumps(cfg.TRAIN)
    print(json.dumps(cfg.TRAIN, sort_keys=False, indent=4))

    print(json.dumps(cfg.TRAIN))



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-16 14:15 
Written by Yuwei Jin (642281525@qq.com)
"""

import torch.nn as nn
import numpy as np
import torch
import copy


# from se_block import SEBlock


def conv_bn(in_channels, out_channels, kernel_size, stride, padding, groups=1):
    result = nn.Sequential()
    result.add_module('conv', nn.Conv2d(in_channels=in_channels, out_channels=out_channels,
                                        kernel_size=kernel_size, stride=stride, padding=padding, groups=groups, bias=False))
    result.add_module('bn', nn.BatchNorm2d(num_features=out_channels))
    return result


class RepVGGBlock(nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size,
                 stride=1, padding=0, dilation=1, groups=1, padding_mode='zeros', deploy=False, use_se=False):
        super(RepVGGBlock, self).__init__()
        self.deploy = deploy
        self.groups = groups
        self.in_channels = in_channels

        assert kernel_size == 3
        assert padding == 1

        padding_11 = padding - kernel_size // 2

        self.nonlinearity = nn.ReLU()

        if use_se:
            pass
            # self.se = SEBlock(out_channels, internal_neurons=out_channels // 16)
        else:
            self.se = nn.Identity()

        if deploy:
            self.rbr_reparam = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride,
                                         padding=padding, dilation=dilation, groups=groups, bias=True, padding_mode=padding_mode)

        else:
            self.rbr_identity = nn.BatchNorm2d(num_features=in_channels) if out_channels == in_channels and stride == 1 else None
            self.rbr_dense = conv_bn(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding,
                                     groups=groups)
            self.rbr_1x1 = conv_bn(in_channels=in_channels, out_channels=out_channels, kernel_size=1, stride=stride, padding=padding_11,
                                   groups=groups)
            print('RepVGG Block, identity = ', self.rbr_identity)

    def forward(self, inputs):
        if hasattr(self, 'rbr_reparam'):
            return self.nonlinearity(self.se(self.rbr_reparam(inputs)))

        if self.rbr_identity is None:
            id_out = 0
        else:
            id_out = self.rbr_identity(inputs)

        return self.nonlinearity(self.se(self.rbr_dense(inputs) + self.rbr_1x1(inputs) + id_out))

    #   Optional. This improves the accuracy and facilitates quantization.
    #   1.  Cancel the original weight decay on rbr_dense.conv.weight and rbr_1x1.conv.weight.
    #   2.  Use like this.
    #       loss = criterion(....)
    #       for every RepVGGBlock blk:
    #           loss += weight_decay_coefficient * 0.5 * blk.get_cust_L2()
    #       optimizer.zero_grad()
    #       loss.backward()
    def get_custom_L2(self):
        K3 = self.rbr_dense.conv.weight
        K1 = self.rbr_1x1.conv.weight
        t3 = (self.rbr_dense.bn.weight / ((self.rbr_dense.bn.running_var + self.rbr_dense.bn.eps).sqrt())).reshape(-1, 1, 1, 1).detach()
        t1 = (self.rbr_1x1.bn.weight / ((self.rbr_1x1.bn.running_var + self.rbr_1x1.bn.eps).sqrt())).reshape(-1, 1, 1, 1).detach()

        l2_loss_circle = (K3 ** 2).sum() - (
                K3[:, :, 1:2, 1:2] ** 2).sum()  # The L2 loss of the "circle" of weights in 3x3 kernel. Use regular L2 on them.
        eq_kernel = K3[:, :, 1:2, 1:2] * t3 + K1 * t1  # The equivalent resultant central point of 3x3 kernel.
        l2_loss_eq_kernel = (eq_kernel ** 2 / (t3 ** 2 + t1 ** 2)).sum()  # Normalize for an L2 coefficient comparable to regular L2.
        return l2_loss_eq_kernel + l2_loss_circle

    #   This func derives the equivalent kernel and bias in a DIFFERENTIABLE way.
    #   You can get the equivalent kernel and bias at any time and do whatever you want,
    #   for example, apply some penalties or constraints during training, just like you do to the other models.
    #   May be useful for quantization or pruning.
    def get_equivalent_kernel_bias(self):
        kernel3x3, bias3x3 = self._fuse_bn_tensor(self.rbr_dense)
        kernel1x1, bias1x1 = self._fuse_bn_tensor(self.rbr_1x1)
        kernelid, biasid = self._fuse_bn_tensor(self.rbr_identity)
        return kernel3x3 + self._pad_1x1_to_3x3_tensor(kernel1x1) + kernelid, bias3x3 + bias1x1 + biasid

    def _pad_1x1_to_3x3_tensor(self, kernel1x1):
        if kernel1x1 is None:
            return 0
        else:
            return torch.nn.functional.pad(kernel1x1, [1, 1, 1, 1])

    def _fuse_bn_tensor(self, branch):
        if branch is None:
            return 0, 0
        if isinstance(branch, nn.Sequential):
            kernel = branch.conv.weight
            running_mean = branch.bn.running_mean
            running_var = branch.bn.running_var
            gamma = branch.bn.weight
            beta = branch.bn.bias
            eps = branch.bn.eps
        else:
            assert isinstance(branch, nn.BatchNorm2d)
            if not hasattr(self, 'id_tensor'):
                input_dim = self.in_channels // self.groups
                kernel_value = np.zeros((self.in_channels, input_dim, 3, 3), dtype=np.float32)
                for i in range(self.in_channels):
                    kernel_value[i, i % input_dim, 1, 1] = 1
                self.id_tensor = torch.from_numpy(kernel_value).to(branch.weight.device)
            kernel = self.id_tensor
            running_mean = branch.running_mean
            running_var = branch.running_var
            gamma = branch.weight
            beta = branch.bias
            eps = branch.eps
        std = (running_var + eps).sqrt()
        t = (gamma / std).reshape(-1, 1, 1, 1)
        return kernel * t, beta - running_mean * gamma / std

    def switch_to_deploy(self):
        if hasattr(self, 'rbr_reparam'):
            return
        kernel, bias = self.get_equivalent_kernel_bias()
        self.rbr_reparam = nn.Conv2d(in_channels=self.rbr_dense.conv.in_channels, out_channels=self.rbr_dense.conv.out_channels,
                                     kernel_size=self.rbr_dense.conv.kernel_size, stride=self.rbr_dense.conv.stride,
                                     padding=self.rbr_dense.conv.padding, dilation=self.rbr_dense.conv.dilation, groups=self.rbr_dense.conv.groups,
                                     bias=True)
        self.rbr_reparam.weight.data = kernel
        self.rbr_reparam.bias.data = bias
        for para in self.parameters():
            para.detach_()
        self.__delattr__('rbr_dense')
        self.__delattr__('rbr_1x1')
        if hasattr(self, 'rbr_identity'):
            self.__delattr__('rbr_identity')


class RepVGG(nn.Module):

    def __init__(self, num_blocks, num_classes=1000, width_multiplier=None, override_groups_map=None, deploy=False, use_se=False):
        super(RepVGG, self).__init__()

        assert len(width_multiplier) == 4

        self.deploy = deploy
        self.override_groups_map = override_groups_map or dict()
        self.use_se = use_se

        assert 0 not in self.override_groups_map

        self.in_planes = min(64, int(64 * width_multiplier[0]))

        self.stage0 = RepVGGBlock(in_channels=3, out_channels=self.in_planes, kernel_size=3, stride=2, padding=1, deploy=self.deploy,
                                  use_se=self.use_se)
        self.cur_layer_idx = 1
        self.stage1 = self._make_stage(int(64 * width_multiplier[0]), num_blocks[0], stride=2)
        self.stage2 = self._make_stage(int(128 * width_multiplier[1]), num_blocks[1], stride=2)
        self.stage3 = self._make_stage(int(256 * width_multiplier[2]), num_blocks[2], stride=2)
        self.stage4 = self._make_stage(int(512 * width_multiplier[3]), num_blocks[3], stride=2)
        self.gap = nn.AdaptiveAvgPool2d(output_size=1)
        self.linear = nn.Linear(int(512 * width_multiplier[3]), num_classes)

    def _make_stage(self, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        blocks = []
        for stride in strides:
            cur_groups = self.override_groups_map.get(self.cur_layer_idx, 1)
            blocks.append(RepVGGBlock(in_channels=self.in_planes, out_channels=planes, kernel_size=3,
                                      stride=stride, padding=1, groups=cur_groups, deploy=self.deploy, use_se=self.use_se))
            self.in_planes = planes
            self.cur_layer_idx += 1
        return nn.Sequential(*blocks)

    def forward(self, x):
        out0 = self.stage0(x)
        out1 = self.stage1(out0)
        out2 = self.stage2(out1)
        out3 = self.stage3(out2)
        out4 = self.stage4(out3)
        # out = self.gap(out)
        # out = out.view(out.size(0), -1)
        # out = self.linear(out)
        return out1, out2, out3, out4


optional_groupwise_layers = [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26]
g2_map = {l: 2 for l in optional_groupwise_layers}
g4_map = {l: 4 for l in optional_groupwise_layers}


def create_RepVGG_A0(deploy=False, pretrained=False, ckpt_src=None):
    train_model = RepVGG(num_blocks=[6, 4, 6, 1], num_classes=1000,
                         width_multiplier=[0.75, 0.75, 0.75, 2.5], override_groups_map=False, deploy=deploy)
    if pretrained:
        train_model.load_state_dict(torch.load(ckpt_src))  # or train from
    return train_model


func_dict = {
    'RepVGG-A0': create_RepVGG_A0  # Updated at April 25, 2021. This is not reported in the CVPR paper.
}


def get_RepVGG_func_by_name(name):
    return func_dict[name]


def repvgg_model_convert(model: torch.nn.Module, save_path=None, do_copy=True):
    if do_copy:
        model = copy.deepcopy(model)
    for module in model.modules():
        if hasattr(module, 'switch_to_deploy'):
            module.switch_to_deploy()
    if save_path is not None:
        torch.save(model.state_dict(), save_path)
    return model


if __name__ == '__main__':
    model = create_RepVGG_A0(deploy=False)
    # del model.linear
    # del model.gap

    from torchstat import stat

    stat(model, (3, 512, 512))

    # train_model = create_RepVGG_A0(deploy=False)
    # train_model.load_state_dict(torch.load('RepVGG-A0-train.pth'))  # or train from scratch
    # deploy_model = repvgg_model_convert(train_model, save_path='RepVGG-A0-deploy.pth')
    #
    # from torchstat import stat
    #
    # stat(deploy_model, (3, 512, 512))



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-23 15:01 
Written by Yuwei Jin (642281525@qq.com)
"""
import torch
import torch.nn as nn
from torch.nn import init
import math


class ConvX(nn.Module):
    def __init__(self, in_planes, out_planes, kernel=3, stride=1):
        super(ConvX, self).__init__()
        self.conv = nn.Conv2d(in_planes, out_planes, kernel_size=kernel, stride=stride, padding=kernel // 2, bias=False)
        self.bn = nn.BatchNorm2d(out_planes)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        out = self.relu(self.bn(self.conv(x)))
        return out


class AddBottleneck(nn.Module):
    def __init__(self, in_planes, out_planes, block_num=3, stride=1):
        super(AddBottleneck, self).__init__()
        assert block_num > 1, print("block number should be larger than 1.")
        self.conv_list = nn.ModuleList()
        self.stride = stride
        if stride == 2:
            self.avd_layer = nn.Sequential(
                nn.Conv2d(out_planes // 2, out_planes // 2, kernel_size=3, stride=2, padding=1, groups=out_planes // 2, bias=False),
                nn.BatchNorm2d(out_planes // 2),
            )
            self.skip = nn.Sequential(
                nn.Conv2d(in_planes, in_planes, kernel_size=3, stride=2, padding=1, groups=in_planes, bias=False),
                nn.BatchNorm2d(in_planes),
                nn.Conv2d(in_planes, out_planes, kernel_size=1, bias=False),
                nn.BatchNorm2d(out_planes),
            )
            stride = 1

        for idx in range(block_num):
            if idx == 0:
                self.conv_list.append(ConvX(in_planes, out_planes // 2, kernel=1))
            elif idx == 1 and block_num == 2:
                self.conv_list.append(ConvX(out_planes // 2, out_planes // 2, stride=stride))
            elif idx == 1 and block_num > 2:
                self.conv_list.append(ConvX(out_planes // 2, out_planes // 4, stride=stride))
            elif idx < block_num - 1:
                self.conv_list.append(ConvX(out_planes // int(math.pow(2, idx)), out_planes // int(math.pow(2, idx + 1))))
            else:
                self.conv_list.append(ConvX(out_planes // int(math.pow(2, idx)), out_planes // int(math.pow(2, idx))))

    def forward(self, x):
        out_list = []
        out = x

        for idx, conv in enumerate(self.conv_list):
            if idx == 0 and self.stride == 2:
                out = self.avd_layer(conv(out))
            else:
                out = conv(out)
            out_list.append(out)

        if self.stride == 2:
            x = self.skip(x)

        return torch.cat(out_list, dim=1) + x


class CatBottleneck(nn.Module):
    def __init__(self, in_planes, out_planes, block_num=3, stride=1):
        super(CatBottleneck, self).__init__()
        assert block_num > 1, print("block number should be larger than 1.")
        self.conv_list = nn.ModuleList()
        self.stride = stride
        if stride == 2:
            self.avd_layer = nn.Sequential(
                nn.Conv2d(out_planes // 2, out_planes // 2, kernel_size=3, stride=2, padding=1, groups=out_planes // 2, bias=False),
                nn.BatchNorm2d(out_planes // 2),
            )
            self.skip = nn.AvgPool2d(kernel_size=3, stride=2, padding=1)
            stride = 1

        for idx in range(block_num):
            if idx == 0:
                self.conv_list.append(ConvX(in_planes, out_planes // 2, kernel=1))
            elif idx == 1 and block_num == 2:
                self.conv_list.append(ConvX(out_planes // 2, out_planes // 2, stride=stride))
            elif idx == 1 and block_num > 2:
                self.conv_list.append(ConvX(out_planes // 2, out_planes // 4, stride=stride))
            elif idx < block_num - 1:
                self.conv_list.append(ConvX(out_planes // int(math.pow(2, idx)), out_planes // int(math.pow(2, idx + 1))))
            else:
                self.conv_list.append(ConvX(out_planes // int(math.pow(2, idx)), out_planes // int(math.pow(2, idx))))

    def forward(self, x):
        out_list = []
        out1 = self.conv_list[0](x)

        for idx, conv in enumerate(self.conv_list[1:]):
            if idx == 0:
                if self.stride == 2:
                    out = conv(self.avd_layer(out1))
                else:
                    out = conv(out1)
            else:
                out = conv(out)
            out_list.append(out)

        if self.stride == 2:
            out1 = self.skip(out1)
        out_list.insert(0, out1)

        out = torch.cat(out_list, dim=1)
        return out


# STDC2Net
class STDCNet1446(nn.Module):
    def __init__(self, base=64, layers=[4, 5, 3], block_num=4, type="cat", num_classes=1000, dropout=0.20, pretrain_model='', use_conv_last=False):
        super(STDCNet1446, self).__init__()
        if type == "cat":
            block = CatBottleneck
        elif type == "add":
            block = AddBottleneck
        self.use_conv_last = use_conv_last
        self.features = self._make_layers(base, layers, block_num, block)
        self.conv_last = ConvX(base * 16, max(1024, base * 16), 1, 1)
        self.gap = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Linear(max(1024, base * 16), max(1024, base * 16), bias=False)
        self.bn = nn.BatchNorm1d(max(1024, base * 16))
        self.relu = nn.ReLU(inplace=True)
        self.dropout = nn.Dropout(p=dropout)
        self.linear = nn.Linear(max(1024, base * 16), num_classes, bias=False)

        self.x2 = nn.Sequential(self.features[:1])
        self.x4 = nn.Sequential(self.features[1:2])
        self.x8 = nn.Sequential(self.features[2:6])
        self.x16 = nn.Sequential(self.features[6:11])
        self.x32 = nn.Sequential(self.features[11:])

        if pretrain_model:
            print('use pretrain model {}'.format(pretrain_model))
            self.init_weight(pretrain_model)
        else:
            self.init_params()

    def init_weight(self, pretrain_model):

        state_dict = torch.load(pretrain_model)["state_dict"]
        self_state_dict = self.state_dict()
        for k, v in state_dict.items():
            self_state_dict.update({k: v})
        self.load_state_dict(self_state_dict)

    def init_params(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def _make_layers(self, base, layers, block_num, block):
        features = []
        features += [ConvX(3, base // 2, 3, 2)]
        features += [ConvX(base // 2, base, 3, 2)]

        for i, layer in enumerate(layers):
            for j in range(layer):
                if i == 0 and j == 0:
                    features.append(block(base, base * 4, block_num, 2))
                elif j == 0:
                    features.append(block(base * int(math.pow(2, i + 1)), base * int(math.pow(2, i + 2)), block_num, 2))
                else:
                    features.append(block(base * int(math.pow(2, i + 2)), base * int(math.pow(2, i + 2)), block_num, 1))

        return nn.Sequential(*features)

    def forward(self, x):
        feat2 = self.x2(x)
        feat4 = self.x4(feat2)
        feat8 = self.x8(feat4)
        feat16 = self.x16(feat8)
        feat32 = self.x32(feat16)
        if self.use_conv_last:
            feat32 = self.conv_last(feat32)

        return feat4, feat8, feat16, feat32

    def forward_impl(self, x):
        out = self.features(x)
        out = self.conv_last(out).pow(2)
        out = self.gap(out).flatten(1)
        out = self.fc(out)
        # out = self.bn(out)
        out = self.relu(out)
        # out = self.relu(self.bn(self.fc(out)))
        out = self.dropout(out)
        out = self.linear(out)
        return out


# STDC1Net
class STDCNet813(nn.Module):
    def __init__(self, base=64, layers=[2, 2, 2], block_num=4, type="cat", num_classes=1000, dropout=0.20, pretrain_model='', use_conv_last=False):
        super(STDCNet813, self).__init__()
        if type == "cat":
            block = CatBottleneck
        elif type == "add":
            block = AddBottleneck
        self.use_conv_last = use_conv_last
        self.features = self._make_layers(base, layers, block_num, block)
        self.conv_last = ConvX(base * 16, max(1024, base * 16), 1, 1)
        self.gap = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Linear(max(1024, base * 16), max(1024, base * 16), bias=False)
        self.bn = nn.BatchNorm1d(max(1024, base * 16))
        self.relu = nn.ReLU(inplace=True)
        self.dropout = nn.Dropout(p=dropout)
        self.linear = nn.Linear(max(1024, base * 16), num_classes, bias=False)

        self.x2 = nn.Sequential(self.features[:1])
        self.x4 = nn.Sequential(self.features[1:2])
        self.x8 = nn.Sequential(self.features[2:4])
        self.x16 = nn.Sequential(self.features[4:6])
        self.x32 = nn.Sequential(self.features[6:])

        if pretrain_model:
            print('use pretrain model {}'.format(pretrain_model))
            self.init_weight(pretrain_model)
        else:
            self.init_params()

    def init_weight(self, pretrain_model):

        state_dict = torch.load(pretrain_model)["state_dict"]
        self_state_dict = self.state_dict()
        for k, v in state_dict.items():
            self_state_dict.update({k: v})

        self.load_state_dict(self_state_dict)

    def init_params(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, std=0.001)
                if m.bias is not None:
                    init.constant_(m.bias, 0)

    def _make_layers(self, base, layers, block_num, block):
        features = []
        features += [ConvX(3, base // 2, 3, 2)]
        features += [ConvX(base // 2, base, 3, 2)]

        for i, layer in enumerate(layers):
            for j in range(layer):
                if i == 0 and j == 0:
                    features.append(block(base, base * 4, block_num, 2))
                elif j == 0:
                    features.append(block(base * int(math.pow(2, i + 1)), base * int(math.pow(2, i + 2)), block_num, 2))
                else:
                    features.append(block(base * int(math.pow(2, i + 2)), base * int(math.pow(2, i + 2)), block_num, 1))

        return nn.Sequential(*features)

    def forward(self, x):
        feat2 = self.x2(x)
        feat4 = self.x4(feat2)
        feat8 = self.x8(feat4)
        feat16 = self.x16(feat8)
        feat32 = self.x32(feat16)
        if self.use_conv_last:
            feat32 = self.conv_last(feat32)

        return feat4, feat8, feat16, feat32

    def forward_impl(self, x):
        out = self.features(x)
        out = self.conv_last(out).pow(2)
        out = self.gap(out).flatten(1)
        out = self.fc(out)
        # out = self.bn(out)
        out = self.relu(out)
        # out = self.relu(self.bn(self.fc(out)))
        out = self.dropout(out)
        out = self.linear(out)
        return out


def stdcnet14446(ckpt_src=None):
    model = STDCNet1446(num_classes=1000, dropout=0.00, block_num=4)
    ckpt = torch.load(ckpt_src, map_location='cpu')
    state_dict = ckpt["state_dict"]
    self_state_dict = model.state_dict()
    for k, v in state_dict.items():
        self_state_dict.update({k: v})
    model.load_state_dict(self_state_dict)
    return model


def stdcnet813(ckpt_src=None):
    model = STDCNet813(num_classes=1000, dropout=0.00, block_num=4)
    ckpt = torch.load(ckpt_src, map_location='cpu')
    state_dict = ckpt["state_dict"]
    self_state_dict = model.state_dict()
    for k, v in state_dict.items():
        self_state_dict.update({k: v})
    model.load_state_dict(self_state_dict)
    return model



if __name__ == "__main__":
    model = stdcnet813()
    model.eval()
    x = torch.randn(1, 3, 224, 224)
    # y = model(x)
    # torch.save(model.state_dict(), 'cat.pth')
    # print(y.size())
    # ckpt = torch.load(r'E:\Building Extraction\big_models\libs\seg_backbone\STDCNet\STDCNet1446_76_47.pth', map_location='cpu')
    # model.load_state_dict( ckpt["state_dict"], strict=False)  # or train from scratch
    from torchstat import stat
    stat(model, (3, 512, 512))